import os
import sys
import re
import socket
import traceback

from six import iteritems

from pandajedi.jedicore.MsgWrapper import MsgWrapper
from .TypicalWatchDogBase import TypicalWatchDogBase
from pandaserver.dataservice.Activator import Activator

# logger
from pandacommon.pandalogger.PandaLogger import PandaLogger
logger = PandaLogger().getLogger(__name__.split('.')[-1])


# watchdog for ATLAS analysis
class AtlasAnalWatchDog(TypicalWatchDogBase):

    # constructor
    def __init__(self, taskBufferIF, ddmIF):
        TypicalWatchDogBase.__init__(self, taskBufferIF, ddmIF)
        self.pid = '{0}-{1}-dog'.format(socket.getfqdn().split('.')[0], os.getpid())
        # self.cronActions = {'forPrestage': 'atlas_prs'}

    # main
    def doAction(self):
        try:
            # get logger
            origTmpLog = MsgWrapper(logger)
            origTmpLog.debug('start')
            # handle waiting jobs
            self.doForWaitingJobs()
            # throttle tasks if so many prestaging requests
            self.doForPreStaging()
            # priority massage
            self.doForPriorityMassage()
            # redo stalled analysis jobs
            self.doForRedoStalledJobs()
            # throttle WAN data access
            #self.doForThrottleWAN()
        except Exception:
            errtype,errvalue = sys.exc_info()[:2]
            origTmpLog.error('failed with {0} {1}'.format(errtype,errvalue))
        # return
        origTmpLog.debug('done')
        return self.SC_SUCCEEDED


    # handle waiting jobs
    def doForWaitingJobs(self):
        tmpLog = MsgWrapper(logger, 'doForWaitingJobs label=user')
        # check every 60 min
        checkInterval = 60
        # get lib.tgz for waiting jobs
        libList = self.taskBufferIF.getLibForWaitingRunJob_JEDI(self.vo, self.prodSourceLabel, checkInterval)
        tmpLog.debug('got {0} lib.tgz files'.format(len(libList)))
        # activate or kill orphan jobs which were submitted to use lib.tgz when the lib.tgz was being produced
        for prodUserName,datasetName,tmpFileSpec in libList:
            tmpLog = MsgWrapper(logger,'< #ATM #KV doForWaitingJobs jediTaskID={0} label=user >'.format(tmpFileSpec.jediTaskID))
            tmpLog.debug('start')
            # check status of lib.tgz
            if tmpFileSpec.status == 'failed':
                # get buildJob
                pandaJobSpecs = self.taskBufferIF.peekJobs([tmpFileSpec.PandaID],
                                                           fromDefined=False,
                                                           fromActive=False,
                                                           fromWaiting=False)
                pandaJobSpec = pandaJobSpecs[0]
                if pandaJobSpec is not None:
                    # kill
                    self.taskBufferIF.updateJobs([pandaJobSpec],False)
                    tmpLog.debug('  action=killed_downstream_jobs for user="{0}" with libDS={1}'.format(prodUserName,datasetName))
                else:
                    # PandaJobSpec not found
                    tmpLog.error('  cannot find PandaJobSpec for user="{0}" with PandaID={1}'.format(prodUserName,
                                                                                                     tmpFileSpec.PandaID))
            elif tmpFileSpec.status == 'finished':
                # set metadata
                self.taskBufferIF.setGUIDs([{'guid':tmpFileSpec.GUID,
                                             'lfn':tmpFileSpec.lfn,
                                             'checksum':tmpFileSpec.checksum,
                                             'fsize':tmpFileSpec.fsize,
                                             'scope':tmpFileSpec.scope,
                                             }])
                # get lib dataset
                dataset = self.taskBufferIF.queryDatasetWithMap({'name':datasetName})
                if dataset is not None:
                    # activate jobs
                    aThr = Activator(self.taskBufferIF,dataset)
                    aThr.start()
                    aThr.join()
                    tmpLog.debug('  action=activated_downstream_jobs for user="{0}" with libDS={1}'.format(prodUserName,datasetName))
                else:
                    # datasetSpec not found
                    tmpLog.error('  cannot find datasetSpec for user="{0}" with libDS={1}'.format(prodUserName,datasetName))
            else:
                # lib.tgz is not ready
                tmpLog.debug('  keep waiting for user="{0}" libDS={1}'.format(prodUserName,datasetName))


    # throttle tasks if so many prestaging requests
    def doForPreStaging(self):
        try:
            tmpLog = MsgWrapper(logger, ' #ATM #KV doForPreStaging label=user')
            tmpLog.debug('start')
            # lock
            got_lock = self.taskBufferIF.lockProcess_JEDI(  vo=self.vo, prodSourceLabel=self.prodSourceLabel,
                                                            cloud=None, workqueue_id=None, resource_name=None,
                                                            component='AtlasAnalWatchDog.doForPreStaging',
                                                            pid=self.pid, timeLimit=5)
            if not got_lock:
                tmpLog.debug('locked by another process. Skipped')
                return
            # get throttled users
            thrUserTasks = self.taskBufferIF.getThrottledUsersTasks_JEDI(self.vo, self.prodSourceLabel)
            # get dispatch datasets
            dispUserTasks = self.taskBufferIF.getDispatchDatasetsPerUser(self.vo, self.prodSourceLabel, True, True)
            # max size of prestaging requests in GB
            maxPrestaging = self.taskBufferIF.getConfigValue('anal_watchdog', 'USER_PRESTAGE_LIMIT', 'jedi', 'atlas')
            if maxPrestaging is None:
                maxPrestaging = 1024
            # max size of transfer requests in GB
            maxTransfer = self.taskBufferIF.getConfigValue('anal_watchdog', 'USER_TRANSFER_LIMIT', 'jedi', 'atlas')
            if maxTransfer is None:
                maxTransfer = 1024
            # throttle interval
            thrInterval = 120
            # loop over all users
            for userName, userDict in iteritems(dispUserTasks):
                # loop over all transfer types
                for transferType, maxSize in [('prestaging', maxPrestaging),
                                               ('transfer', maxTransfer)]:
                    if transferType not in userDict:
                        continue
                    userTotal = userDict[transferType]['size'] / 1024
                    tmpLog.debug('user={0} {1} total={2} GB'.format(userName, transferType, userTotal))
                    # too large
                    if userTotal > maxSize:
                        tmpLog.debug('user={0} has too large {1} total={2} GB > limit={3} GB'.
                                     format(userName, transferType, userTotal, maxSize))
                        # throttle tasks
                        for taskID in userDict[transferType]['tasks']:
                            if userName not in thrUserTasks or transferType not in thrUserTasks[userName] \
                                    or taskID not in thrUserTasks[userName][transferType]:
                                tmpLog.debug('action=throttle_{0} jediTaskID={1} for user={2}'.format(transferType,
                                                                                                      taskID, userName))
                                errDiag = 'throttled for {0} min due to large data motion type={1}'.format(thrInterval,
                                                                                                           transferType)
                                self.taskBufferIF.throttleTask_JEDI(taskID,thrInterval,errDiag)
                        # remove the user from the list
                        if userName in thrUserTasks and transferType in thrUserTasks[userName]:
                            del thrUserTasks[userName][transferType]
            # release users
            for userName, taskData in iteritems(thrUserTasks):
                for transferType, taskIDs in iteritems(taskData):
                    tmpLog.debug('user={0} release throttled tasks with {1}'.format(userName, transferType))
                    # unthrottle tasks
                    for taskID in taskIDs:
                        tmpLog.debug('action=release_{0} jediTaskID={1} for user={2}'.format(transferType,
                                                                                             taskID, userName))
                        self.taskBufferIF.releaseThrottledTask_JEDI(taskID)
            tmpLog.debug('done')
        except Exception:
            errtype, errvalue = sys.exc_info()[:2]
            tmpLog.error('failed with {0} {1} {2}'.format(errtype, errvalue, traceback.format_exc()))


    # priority massage
    def doForPriorityMassage(self):
        tmpLog = MsgWrapper(logger, ' #ATM #KV doForPriorityMassage label=user')
        tmpLog.debug('start')
        # lock
        got_lock = self.taskBufferIF.lockProcess_JEDI(  vo=self.vo, prodSourceLabel=self.prodSourceLabel,
                                                        cloud=None, workqueue_id=None, resource_name=None,
                                                        component='AtlasAnalWatchDog.doForPriorityMassage',
                                                        pid=self.pid, timeLimit=6)
        if not got_lock:
            tmpLog.debug('locked by another process. Skipped')
            return
        try:
            # get usage breakdown
            usageBreakDownPerUser, usageBreakDownPerSite = self.taskBufferIF.getUsageBreakdown_JEDI(self.prodSourceLabel)
            # get total number of users and running/done jobs
            totalUsers = 0
            totalRunDone = 0
            usersTotalJobs = {}
            usersTotalCores = {}
            for prodUserName in usageBreakDownPerUser:
                wgValMap = usageBreakDownPerUser[prodUserName]
                for workingGroup in wgValMap:
                    siteValMap = wgValMap[workingGroup]
                    totalUsers += 1
                    for computingSite in siteValMap:
                        statValMap = siteValMap[computingSite]
                        totalRunDone += statValMap['rundone']
                        usersTotalJobs.setdefault(prodUserName, {})
                        usersTotalJobs[prodUserName].setdefault(workingGroup, 0)
                        usersTotalJobs[prodUserName][workingGroup] += statValMap['running']
                        usersTotalCores.setdefault(prodUserName, {})
                        usersTotalCores[prodUserName].setdefault(workingGroup, 0)
                        usersTotalCores[prodUserName][workingGroup] += statValMap['runcores']
            tmpLog.debug('total {0} users, {1} RunDone jobs'.format(totalUsers, totalRunDone))
            # skip if no user
            if totalUsers == 0:
                tmpLog.debug('no user. Skipped...')
                return
            # cap num of running jobs
            tmpLog.debug('cap running jobs')
            prodUserName = None
            maxNumRunPerUser = self.taskBufferIF.getConfigValue('prio_mgr', 'CAP_RUNNING_USER_JOBS')
            maxNumRunPerGroup = self.taskBufferIF.getConfigValue('prio_mgr', 'CAP_RUNNING_GROUP_JOBS')
            maxNumCorePerUser = self.taskBufferIF.getConfigValue('prio_mgr', 'CAP_RUNNING_USER_CORES')
            maxNumCorePerGroup = self.taskBufferIF.getConfigValue('prio_mgr', 'CAP_RUNNING_GROUP_CORES')
            if maxNumRunPerUser is None:
                maxNumRunPerUser = 10000
            if maxNumRunPerGroup is None:
                maxNumRunPerGroup = 10000
            if maxNumCorePerUser is None:
                maxNumCorePerUser = 10000
            if maxNumCorePerGroup is None:
                maxNumCorePerGroup = 10000
            try:
                throttledUsers = self.taskBufferIF.getThrottledUsers()
                for prodUserName in usersTotalJobs:
                    for workingGroup in usersTotalJobs[prodUserName]:
                        tmpNumTotalJobs = usersTotalJobs[prodUserName][workingGroup]
                        tmpNumTotalCores = usersTotalCores[prodUserName][workingGroup]
                        if workingGroup is None:
                            maxNumRun = maxNumRunPerUser
                            maxNumCore = maxNumCorePerUser
                        else:
                            maxNumRun = maxNumRunPerGroup
                            maxNumCore = maxNumCorePerGroup
                        if tmpNumTotalJobs >= maxNumRun or tmpNumTotalCores >= maxNumCore:
                            # throttle user
                            tmpNumJobs = self.taskBufferIF.throttleUserJobs(prodUserName, workingGroup, get_dict=True)
                            if tmpNumJobs is not None:
                                for tmpJediTaskID, tmpNumJob in iteritems(tmpNumJobs):
                                    msg = ('throttled {} jobs in jediTaskID={} for user="{}" group={} '
                                           'since too many running jobs ({} > {}) or cores ({} > {}) ').format(
                                           tmpNumJob, tmpJediTaskID, prodUserName, workingGroup, tmpNumTotalJobs,
                                           maxNumRun, tmpNumTotalCores, maxNumCore)
                                    tmpLog.debug(msg)
                                    tmpLog.sendMsg(msg, 'userCap', msgLevel='warning')
                        elif tmpNumTotalJobs < maxNumRun*0.9 and tmpNumTotalCores < maxNumCore*0.9 and \
                                (prodUserName, workingGroup) in throttledUsers:
                            # unthrottle user
                            tmpNumJobs = self.taskBufferIF.unThrottleUserJobs(prodUserName, workingGroup, get_dict=True)
                            if tmpNumJobs is not None:
                                for tmpJediTaskID, tmpNumJob in iteritems(tmpNumJobs):
                                    msg = ('released {} jobs in jediTaskID={} for user="{}" group={} '
                                           'since number of running jobs and cores are less than {} and {}').format(
                                           tmpNumJob, tmpJediTaskID, prodUserName, workingGroup, maxNumRun, maxNumCore)
                                    tmpLog.debug(msg)
                                    tmpLog.sendMsg(msg, 'userCap')
            except Exception as e:
                errStr = "cap failed for %s : %s" % (prodUserName, str(e))
                errStr.strip()
                errStr += traceback.format_exc()
                tmpLog.error(errStr)
            # to boost
            tmpLog.debug('boost jobs')
            # global average
            globalAverageRunDone = float(totalRunDone)/float(totalUsers)
            tmpLog.debug('global average: {0}'.format(globalAverageRunDone))
            # count the number of users and run/done jobs for each site
            siteRunDone = {}
            siteUsers = {}
            for computingSite in usageBreakDownPerSite:
                userValMap = usageBreakDownPerSite[computingSite]
                for prodUserName in userValMap:
                    wgValMap = userValMap[prodUserName]
                    for workingGroup in wgValMap:
                        statValMap = wgValMap[workingGroup]
                        # count the number of users and running/done jobs
                        siteUsers.setdefault(computingSite, 0)
                        siteUsers[computingSite] += 1
                        siteRunDone.setdefault(computingSite, 0)
                        siteRunDone[computingSite] += statValMap['rundone']
            # get site average
            tmpLog.debug('site average')
            siteAverageRunDone = {}
            for computingSite in siteRunDone:
                nRunDone = siteRunDone[computingSite]
                siteAverageRunDone[computingSite] = float(nRunDone)/float(siteUsers[computingSite])
                tmpLog.debug(" %-25s : %s" % (computingSite,siteAverageRunDone[computingSite]))
            # check if the number of user's jobs is lower than the average
            for prodUserName in usageBreakDownPerUser:
                wgValMap = usageBreakDownPerUser[prodUserName]
                for workingGroup in wgValMap:
                    tmpLog.debug("---> %s group=%s" % (prodUserName, workingGroup))
                    # count the number of running/done jobs
                    userTotalRunDone = 0
                    for computingSite in wgValMap[workingGroup]:
                        statValMap = wgValMap[workingGroup][computingSite]
                        userTotalRunDone += statValMap['rundone']
                    # no priority boost when the number of jobs is higher than the average
                    if userTotalRunDone >= globalAverageRunDone:
                        tmpLog.debug("enough running %s > %s (global average)" % (userTotalRunDone,globalAverageRunDone))
                        continue
                    tmpLog.debug("user total:%s global average:%s" % (userTotalRunDone,globalAverageRunDone))
                    # check with site average
                    toBeBoostedSites = []
                    for computingSite in wgValMap[workingGroup]:
                        statValMap = wgValMap[workingGroup][computingSite]
                        # the number of running/done jobs is lower than the average and activated jobs are waiting
                        if statValMap['rundone'] >= siteAverageRunDone[computingSite]:
                            tmpLog.debug("enough running %s > %s (site average) at %s" % \
                                          (statValMap['rundone'],siteAverageRunDone[computingSite],computingSite))
                        elif statValMap['activated'] == 0:
                            tmpLog.debug("no activated jobs at %s" % computingSite)
                        else:
                            toBeBoostedSites.append(computingSite)
                    # no boost is required
                    if toBeBoostedSites == []:
                        tmpLog.debug("no sites to be boosted")
                        continue
                    # check special prioritized site
                    siteAccessForUser = {}
                    varMap = {}
                    varMap[':dn'] = prodUserName
                    sql = "SELECT pandaSite,pOffset,status,workingGroups FROM ATLAS_PANDAMETA.siteAccess WHERE dn=:dn"
                    res = self.taskBufferIF.querySQL(sql, varMap, arraySize=10000)
                    if res is not None:
                        for pandaSite, pOffset, pStatus, workingGroups in res:
                            # ignore special working group for now
                            if workingGroups not in ['', None]:
                                continue
                            # only approved sites
                            if pStatus != 'approved':
                                continue
                            # no priority boost
                            if pOffset == 0:
                                continue
                            # append
                            siteAccessForUser[pandaSite] = pOffset
                    # set weight
                    totalW = 0
                    defaultW = 100
                    for computingSite in toBeBoostedSites:
                        totalW += defaultW
                        if computingSite in siteAccessForUser:
                            totalW += siteAccessForUser[computingSite]
                    totalW = float(totalW)
                    # the total number of jobs to be boosted
                    numBoostedJobs = globalAverageRunDone - float(userTotalRunDone)
                    # get quota
                    quotaFactor = 1.0 + self.taskBufferIF.checkQuota(prodUserName)
                    tmpLog.debug("quota factor:%s" % quotaFactor)
                    # make priority boost
                    nJobsPerPrioUnit = 5
                    highestPrio = 1000
                    for computingSite in toBeBoostedSites:
                        weight = float(defaultW)
                        if computingSite in siteAccessForUser:
                            weight += float(siteAccessForUser[computingSite])
                        weight /= totalW
                        # the number of boosted jobs at the site
                        numBoostedJobsSite = int(numBoostedJobs * weight / quotaFactor)
                        tmpLog.debug("nSite:%s nAll:%s W:%s Q:%s at %s" % (numBoostedJobsSite, numBoostedJobs, weight, quotaFactor, computingSite))
                        if numBoostedJobsSite/nJobsPerPrioUnit == 0:
                            tmpLog.debug("too small number of jobs %s to be boosted at %s" % (numBoostedJobsSite, computingSite))
                            continue
                        # get the highest prio of activated jobs at the site
                        varMap = {}
                        varMap[':jobStatus'] = 'activated'
                        varMap[':prodSourceLabel'] = self.prodSourceLabel
                        varMap[':pmerge'] = 'pmerge'
                        varMap[':prodUserName'] = prodUserName
                        varMap[':computingSite'] = computingSite
                        sql  = "SELECT MAX(currentPriority) FROM ATLAS_PANDA.jobsActive4 "
                        sql += "WHERE prodSourceLabel=:prodSourceLabel AND jobStatus=:jobStatus AND computingSite=:computingSite "
                        sql += "AND processingType<>:pmerge AND prodUserName=:prodUserName "
                        if workingGroup is not None:
                            varMap[':workingGroup'] = workingGroup
                            sql += "AND workingGroup=:workingGroup "
                        else:
                            sql += "AND workingGroup IS NULL "
                        res = self.taskBufferIF.querySQL(sql, varMap, arraySize=10)
                        maxPrio = None
                        if res is not None:
                            try:
                                maxPrio = res[0][0]
                            except Exception:
                                pass
                        if maxPrio is None:
                            tmpLog.debug("cannot get the highest prio at %s" % computingSite)
                            continue
                        # delta for priority boost
                        prioDelta = highestPrio - maxPrio
                        # already boosted
                        if prioDelta <= 0:
                            tmpLog.debug("already boosted (prio=%s) at %s" % (maxPrio,computingSite))
                            continue
                        # lower limit
                        minPrio = maxPrio - numBoostedJobsSite/nJobsPerPrioUnit
                        # SQL for priority boost
                        varMap = {}
                        varMap[':jobStatus'] = 'activated'
                        varMap[':prodSourceLabel'] = self.prodSourceLabel
                        varMap[':prodUserName'] = prodUserName
                        varMap[':computingSite'] = computingSite
                        varMap[':prioDelta'] = prioDelta
                        varMap[':maxPrio'] = maxPrio
                        varMap[':minPrio'] = minPrio
                        varMap[':rlimit'] = numBoostedJobsSite
                        sql  = "UPDATE ATLAS_PANDA.jobsActive4 SET currentPriority=currentPriority+:prioDelta "
                        sql += "WHERE prodSourceLabel=:prodSourceLabel AND prodUserName=:prodUserName "
                        if workingGroup is not None:
                            varMap[':workingGroup'] = workingGroup
                            sql += "AND workingGroup=:workingGroup "
                        else:
                            sql += "AND workingGroup IS NULL "
                        sql += "AND jobStatus=:jobStatus AND computingSite=:computingSite AND currentPriority>:minPrio "
                        sql += "AND currentPriority<=:maxPrio AND rownum<=:rlimit"
                        tmpLog.debug("boost %s" % str(varMap))
                        res = self.taskBufferIF.querySQL(sql, varMap, arraySize=10)
                        tmpLog.debug("   database return : %s" % res)
            # done
            tmpLog.debug('done')
        except Exception:
            errtype, errvalue = sys.exc_info()[:2]
            tmpLog.error('failed with {0} {1} {2}'.format(errtype, errvalue, traceback.format_exc()))


    # redo stalled analysis jobs
    def doForRedoStalledJobs(self):
        tmpLog = MsgWrapper(logger, ' #ATM #KV doForRedoStalledJobs label=user')
        tmpLog.debug('start')
        # lock
        got_lock = self.taskBufferIF.lockProcess_JEDI(  vo=self.vo, prodSourceLabel=self.prodSourceLabel,
                                                        cloud=None, workqueue_id=None, resource_name=None,
                                                        component='AtlasAnalWatchDog.doForRedoStalledJobs',
                                                        pid=self.pid, timeLimit=6)
        if not got_lock:
            tmpLog.debug('locked by another process. Skipped')
            return
        # redo stalled analysis jobs
        tmpLog.debug('redo stalled jobs')
        try:
            varMap = {}
            varMap[':prodSourceLabel'] = self.prodSourceLabel
            sqlJ =  "SELECT jobDefinitionID,prodUserName FROM ATLAS_PANDA.jobsDefined4 "
            sqlJ += "WHERE prodSourceLabel=:prodSourceLabel AND modificationTime<CURRENT_DATE-2/24 "
            sqlJ += "GROUP BY jobDefinitionID,prodUserName"
            sqlP  = "SELECT PandaID FROM ATLAS_PANDA.jobsDefined4 "
            sqlP += "WHERE jobDefinitionID=:jobDefinitionID ANd prodSourceLabel=:prodSourceLabel AND prodUserName=:prodUserName AND rownum <= 1"
            sqlF  = "SELECT lfn,type,destinationDBlock FROM ATLAS_PANDA.filesTable4 WHERE PandaID=:PandaID AND status=:status"
            sqlL  = "SELECT guid,status,PandaID,dataset FROM ATLAS_PANDA.filesTable4 WHERE lfn=:lfn AND type=:type"
            sqlA  = "SELECT PandaID FROM ATLAS_PANDA.jobsDefined4 "
            sqlA += "WHERE jobDefinitionID=:jobDefinitionID ANd prodSourceLabel=:prodSourceLabel AND prodUserName=:prodUserName"
            sqlU  = "UPDATE ATLAS_PANDA.jobsDefined4 SET modificationTime=CURRENT_DATE "
            sqlU += "WHERE jobDefinitionID=:jobDefinitionID ANd prodSourceLabel=:prodSourceLabel AND prodUserName=:prodUserName"
            # get stalled jobs
            resJ = self.taskBufferIF.querySQL(sqlJ, varMap)
            if resJ is None or len(resJ) == 0:
                pass
            else:
                # loop over all jobID/users
                for jobDefinitionID,prodUserName in resJ:
                    tmpLog.debug(" user:%s jobID:%s" % (prodUserName,jobDefinitionID))
                    # get stalled jobs
                    varMap = {}
                    varMap[':prodSourceLabel'] = self.prodSourceLabel
                    varMap[':jobDefinitionID'] = jobDefinitionID
                    varMap[':prodUserName']    = prodUserName
                    resP = self.taskBufferIF.querySQL(sqlP, varMap)
                    if resP is None or len(resP) == 0:
                        tmpLog.debug("  no PandaID")
                        continue
                    useLib    = False
                    libStatus = None
                    libGUID   = None
                    libLFN    = None
                    libDSName = None
                    destReady = False
                    # use the first PandaID
                    for PandaID, in resP:
                        tmpLog.debug("  check PandaID:%s" % PandaID)
                        # get files
                        varMap = {}
                        varMap[':PandaID'] = PandaID
                        varMap[':status']  = 'unknown'
                        resF = self.taskBufferIF.querySQL(sqlF, varMap)
                        if resF is None or len(resF) == 0:
                            tmpLog.debug("  no files")
                        else:
                            # get lib.tgz and destDBlock
                            for lfn,filetype,destinationDBlock in resF:
                                if filetype == 'input' and lfn.endswith('.lib.tgz'):
                                    useLib = True
                                    libLFN = lfn
                                    varMap = {}
                                    varMap[':lfn'] = lfn
                                    varMap[':type']  = 'output'
                                    resL = self.taskBufferIF.querySQL(sqlL, varMap)
                                    # not found
                                    if resL is None or len(resL) == 0:
                                        tmpLog.error("  cannot find status of %s" % lfn)
                                        continue
                                    # check status
                                    guid,outFileStatus,pandaIDOutLibTgz,tmpLibDsName = resL[0]
                                    tmpLog.debug("  PandaID:%s produces %s:%s GUID=%s status=%s" % (pandaIDOutLibTgz,tmpLibDsName,lfn,guid,outFileStatus))
                                    libStatus = outFileStatus
                                    libGUID   = guid
                                    libDSName = tmpLibDsName
                                elif filetype in ['log','output']:
                                    if destinationDBlock is not None and re.search('_sub\d+$',destinationDBlock) is not None:
                                        destReady = True
                            break
                    tmpLog.debug("  useLib:%s libStatus:%s libDsName:%s libLFN:%s libGUID:%s destReady:%s" % (useLib,libStatus,libDSName,libLFN,libGUID,destReady))
                    if libStatus == 'failed':
                        # delete downstream jobs
                        tmpLog.debug("  -> delete downstream jobs")
                        # FIXME
                        #self.taskBufferIF.deleteStalledJobs(libLFN)
                    else:
                        # activate
                        if useLib and libStatus == 'ready' and (libGUID not in [None,'']) and (libDSName not in [None,'']):
                            # update GUID
                            tmpLog.debug("  set GUID:%s for %s" % (libGUID,libLFN))
                            #retG = self.taskBufferIF.setGUIDs([{'lfn':libLFN,'guid':libGUID}])
                            # FIXME
                            retG = True
                            if not retG:
                                tmpLog.error("  failed to update GUID for %s" % libLFN)
                            else:
                                # get PandaID with lib.tgz
                                #ids = self.taskBufferIF.updateInFilesReturnPandaIDs(libDSName,'ready')
                                ids = []
                                # get jobs
                                jobs = self.taskBufferIF.peekJobs(ids,fromActive=False,fromArchived=False,fromWaiting=False)
                                # remove None and unknown
                                acJobs = []
                                for job in jobs:
                                    if job is None or job.jobStatus == 'unknown':
                                        continue
                                    acJobs.append(job)
                                # activate
                                tmpLog.debug("  -> activate downstream jobs")
                                #self.taskBufferIF.activateJobs(acJobs)
                        else:
                            # wait
                            tmpLog.debug("  -> wait")
                            varMap = {}
                            varMap[':prodSourceLabel'] = self.prodSourceLabel
                            varMap[':jobDefinitionID'] = jobDefinitionID
                            varMap[':prodUserName']    = prodUserName
                            # FIXME
                            #resU = self.taskBufferIF.querySQL(sqlU, varMap)
            # done
            tmpLog.debug('done')
        except Exception:
            errtype, errvalue = sys.exc_info()[:2]
            tmpLog.error('failed to redo stalled jobs with {0} {1} {2}'.format(errtype, errvalue, traceback.format_exc()))


    # throttle WAN data access
    def doForThrottleWAN(self):
        tmpLog = MsgWrapper(logger, ' #ATM #KV doForThrottleWAN label=user')
        tmpLog.debug('start')
        # lock
        got_lock = self.taskBufferIF.lockProcess_JEDI(  vo=self.vo, prodSourceLabel=self.prodSourceLabel,
                                                        cloud=None, workqueue_id=None, resource_name=None,
                                                        component='AtlasAnalWatchDog.doForThrottleWAN',
                                                        pid=self.pid, timeLimit=6)
        if not got_lock:
            tmpLog.debug('locked by another process. Skipped')
            return
        # throttle WAN data access
        tmpLog.debug('throttle WAN data access')
        try:
            # max number of activated jobs with WAN access
            maxActivated = 5
            # get WAN data matrix
            wanMX = self.taskBufferIF.getWanDataFlowMaxtrix()
            throttleForSink     = {}
            throttleForSource   = {}
            totalFlowFromSource = {}
            # loop over all sources to get total flows
            tmpLog.debug(" >>> checking limits")
            for sinkSite in wanMX:
                sinkMap = wanMX[sinkSite]
                totalFlowToSink = 0
                # loop over all sinks
                for sourceSite in sinkMap:
                    sourceMap = sinkMap[sourceSite]
                    # get total flows
                    totalFlowToSink += sourceMap['flow']
                    totalFlowFromSource.setdefault(sourceSite, 0)
                    totalFlowFromSource[sourceSite] += sourceMap['flow']
                # check limit for sink
                tmpSiteSpec = siteMapper.getSite(sinkSite)
                if siteMapper.checkSite(sinkSite) and tmpSiteSpec.wansinklimit*1024*1024*1024 > totalFlowToSink:
                    throttleForSink[sinkSite] = False
                    tmpLog.debug(" release Sink {0} : {1}bps (total) < {2}Gbps (limit)".format(sinkSite,totalFlowToSink,
                                                            tmpSiteSpec.wansinklimit))
                else:
                    throttleForSink[sinkSite] = True
                    tmpLog.debug(" throttle Sink {0} : {1}bps (total) > {2}Gbps (limit)".format(sinkSite,totalFlowToSink,
                                                             tmpSiteSpec.wansinklimit))
            # check limit for source
            for sourceSite in totalFlowFromSource:
                totalFlow = totalFlowFromSource[sourceSite]
                tmpSiteSpec = siteMapper.getSite(sourceSite)
                if siteMapper.checkSite(sourceSite) and tmpSiteSpec.wansourcelimit*1024*1024*1024 > totalFlow:
                    throttleForSource[sourceSite] = False
                    tmpLog.debug(" release Src {0} : {1}bps (total) < {2}Gbps (limit)".format(sourceSite,totalFlow,
                                                                                              tmpSiteSpec.wansourcelimit))
                else:
                    throttleForSource[sourceSite] = True
                    tmpLog.debug(" throttle Src {0} : {1}bps (total) > {2}Gbps (limit)".format(sourceSite,totalFlow,
                                                            tmpSiteSpec.wansourcelimit))
            # loop over all sources to adjust flows
            tmpLog.debug(" >>> adjusting flows")
            for sinkSite in wanMX:
                sinkMap = wanMX[sinkSite]
                if throttleForSink[sinkSite]:
                    # sink is throttled
                    iJobs = 0
                    for sourceSite in sinkMap:
                        sourceMap = sinkMap[sourceSite]
                        for prodUserName in sourceMap['user']:
                            userMap = sourceMap['user'][prodUserName]
                            for currentPriority in userMap['activated']['jobList']:
                                jobList = userMap['activated']['jobList'][currentPriority]
                                for pandaID in jobList:
                                    tmpStat = self.taskBufferIF.throttleJob(pandaID)
                                    if tmpStat == 1:
                                        iJobs += 1
                    tmpLog.debug(" throttled {0} jobs to {1}".format(iJobs,sinkSite))
                else:
                    # no throttle on sink
                    for sourceSite in sinkMap:
                        sourceMap = sinkMap[sourceSite]
                        # check if source is throttled
                        if throttleForSource[sourceSite]:
                            # throttled
                            iJobs = 0
                            for prodUserName in sourceMap['user']:
                                userMap = sourceMap['user'][prodUserName]
                                for currentPriority in userMap['activated']['jobList']:
                                    jobList = userMap['activated']['jobList'][currentPriority]
                                    for pandaID in jobList:
                                        tmpStat = self.taskBufferIF.throttleJob(pandaID)
                                        if tmpStat == 1:
                                            iJobs += 1
                            tmpLog.debug(" throttled {0} jobs from {1} to {2}".format(iJobs,sourceSite,sinkSite))
                        else:
                            # unthrottled
                            iJobs = 0
                            for prodUserName in sourceMap['user']:
                                userMap = sourceMap['user'][prodUserName]
                                if userMap['activated']['nJobs'] < maxActivated:
                                    # activate jobs with higher priorities
                                    currentPriorityList = list(userMap['throttled']['jobList'])
                                    currentPriorityList.sort()
                                    currentPriorityList.reverse()
                                    nActivated = userMap['activated']['nJobs']
                                    for currentPriority in currentPriorityList:
                                        if nActivated > maxActivated:
                                            break
                                        panaIDs = userMap['throttled']['jobList'][currentPriority]
                                        panaIDs.sort()
                                        for pandaID in panaIDs:
                                            tmpStat = self.taskBufferIF.unThrottleJob(pandaID)
                                            if tmpStat == 1:
                                                nActivated += 1
                                                iJobs += 1
                                            if nActivated > maxActivated:
                                                break
                            tmpLog.debug(" activated {0} jobs from {1} to {2}".format(iJobs,sourceSite,sinkSite))
            # done
            tmpLog.debug('done')
        except Exception:
            errtype,errvalue = sys.exc_info()[:2]
            tmpLog.error("failed to throttle WAN data access with %s %s" % (errtype,errvalue))



from pandajedi.jedicore import Interaction


# base class for watchdog
class WatchDogBase(object):

    # constructor
    def __init__(self, taskBufferIF, ddmIF):
        self.taskBufferIF = taskBufferIF
        self.ddmIF = ddmIF
        self.refresh()

    # refresh
    def refresh(self):
        self.siteMapper = self.taskBufferIF.getSiteMapper()

    # pre-action
    def pre_action(self, tmpLog, vo, prodSourceLabel, pid, *args, **kwargs):
        pass


Interaction.installSC(WatchDogBase)



import os
import re
import sys
import time
import socket
import datetime
import traceback

from six import iteritems

from pandajedi.jedicore import Interaction
from pandajedi.jedicore.ThreadUtils import ListWithLock,ThreadPool,WorkerThread
from pandajedi.jedicore.MsgWrapper import MsgWrapper
from .JediKnight import JediKnight
from pandajedi.jediconfig import jedi_config
from pandajedi.jedicore.JediTaskSpec import JediTaskSpec
from pandajedi.jedirefine import RefinerUtils


# logger
from pandacommon.pandalogger.PandaLogger import PandaLogger
logger = PandaLogger().getLogger(__name__.split('.')[-1])


# worker class to kill/amend tasks
class TaskCommando (JediKnight):

    # constructor
    def __init__(self,commuChannel,taskBufferIF,ddmIF,vos,prodSourceLabels):
        self.vos = self.parseInit(vos)
        self.prodSourceLabels = self.parseInit(prodSourceLabels)
        self.pid = '{0}-{1}-dog'.format(socket.getfqdn().split('.')[0],os.getpid())
        JediKnight.__init__(self,commuChannel,taskBufferIF,ddmIF,logger)


    # main
    def start(self):
        # start base classes
        JediKnight.start(self)
        # go into main loop
        while True:
            startTime = datetime.datetime.utcnow()
            try:
                # get logger
                tmpLog = MsgWrapper(logger)
                tmpLog.debug('start')
                # loop over all vos
                for vo in self.vos:
                    # loop over all sourceLabels
                    for prodSourceLabel in self.prodSourceLabels:
                        # get the list of tasks to exec command
                        tmpList = self.taskBufferIF.getTasksToExecCommand_JEDI(vo,prodSourceLabel)
                        if tmpList is None:
                            # failed
                            tmpLog.error('failed to get the task list for vo={0} label={1}'.format(vo,prodSourceLabel))
                        else:
                            tmpLog.debug('got {0} tasks'.format(len(tmpList)))
                            # put to a locked list
                            taskList = ListWithLock(tmpList)
                            # make thread pool
                            threadPool = ThreadPool()
                            # make workers
                            nWorker = jedi_config.taskrefine.nWorkers
                            for iWorker in range(nWorker):
                                thr = TaskCommandoThread(taskList,threadPool,
                                                         self.taskBufferIF,
                                                         self.ddmIF,
                                                         self.pid)
                                thr.start()
                            # join
                            threadPool.join()
                tmpLog.debug('done')
            except Exception as e:
                tmpLog.error('failed in {}.start() with {} {}'.format(self.__class__.__name__, str(e),
                                                                      traceback.format_exc()))
            # sleep if needed
            loopCycle = jedi_config.tcommando.loopCycle
            timeDelta = datetime.datetime.utcnow() - startTime
            sleepPeriod = loopCycle - timeDelta.seconds
            if sleepPeriod > 0:
                time.sleep(sleepPeriod)
            # randomize cycle
            self.randomSleep(max_val=loopCycle)



# thread for real worker
class TaskCommandoThread (WorkerThread):

    # constructor
    def __init__(self,taskList,threadPool,taskbufferIF,ddmIF,pid):
        # initialize woker with no semaphore
        WorkerThread.__init__(self,None,threadPool,logger)
        # attributres
        self.taskList = taskList
        self.taskBufferIF = taskbufferIF
        self.ddmIF = ddmIF
        self.msgType = 'taskcommando'
        self.pid     = pid



    # main
    def runImpl(self):
        while True:
            try:
                # get a part of list
                nTasks = 10
                taskList = self.taskList.get(nTasks)
                # no more datasets
                if len(taskList) == 0:
                    self.logger.debug('{0} terminating since no more items'.format(self.__class__.__name__))
                    return
                # loop over all tasks
                for jediTaskID,commandMap in taskList:
                    # make logger
                    tmpLog = MsgWrapper(self.logger,' < jediTaskID={0} >'.format(jediTaskID))
                    commandStr = commandMap['command']
                    commentStr = commandMap['comment']
                    oldStatus  = commandMap['oldStatus']
                    tmpLog.info('start for {0}'.format(commandStr))
                    tmpStat = Interaction.SC_SUCCEEDED
                    if commandStr in ['kill','finish','reassign']:
                        tmpMsg = 'executing {0}'.format(commandStr)
                        tmpLog.info(tmpMsg)
                        tmpLog.sendMsg(tmpMsg,self.msgType)
                        # loop twice to see immediate result
                        for iLoop in range(2):
                            # get active PandaIDs to be killed
                            if commandStr == 'reassign' and commentStr is not None and 'soft reassign' in commentStr:
                                pandaIDs = self.taskBufferIF.getQueuedPandaIDsWithTask_JEDI(jediTaskID)
                            elif commandStr == 'reassign' and commentStr is not None and 'nokill reassign' in commentStr:
                                pandaIDs = []
                            else:
                                pandaIDs = self.taskBufferIF.getPandaIDsWithTask_JEDI(jediTaskID,True)
                            if pandaIDs is None:
                                tmpLog.error('failed to get PandaIDs for jediTaskID={0}'.format(jediTaskID))
                                tmpStat = Interaction.SC_FAILED
                            # kill jobs or update task
                            if tmpStat == Interaction.SC_SUCCEEDED:
                                if pandaIDs == []:
                                    # done since no active jobs
                                    tmpMsg = 'completed cleaning jobs'
                                    tmpLog.sendMsg(tmpMsg,self.msgType)
                                    tmpLog.info(tmpMsg)
                                    tmpTaskSpec = JediTaskSpec()
                                    tmpTaskSpec.jediTaskID = jediTaskID
                                    updateTaskStatus = True
                                    if commandStr != 'reassign':
                                        # reset oldStatus
                                        # keep oldStatus for task reassignment since it is reset when actually reassigned
                                        tmpTaskSpec.forceUpdate('oldStatus')
                                    else:
                                        # extract cloud or site
                                        if commentStr is not None:
                                            tmpItems = commentStr.split(':')
                                            if tmpItems[0] == 'cloud':
                                                tmpTaskSpec.cloud = tmpItems[1]
                                            elif tmpItems[0] == 'nucleus':
                                                tmpTaskSpec.nucleus = tmpItems[1]
                                            else:
                                                tmpTaskSpec.site = tmpItems[1]
                                            tmpMsg = 'set {0}={1}'.format(tmpItems[0],tmpItems[1])
                                            tmpLog.sendMsg(tmpMsg,self.msgType)
                                            tmpLog.info(tmpMsg)
                                            # back to oldStatus if necessary
                                            if tmpItems[2] == 'y':
                                                tmpTaskSpec.status = oldStatus
                                                tmpTaskSpec.forceUpdate('oldStatus')
                                                updateTaskStatus = False
                                    if commandStr == 'reassign':
                                        tmpTaskSpec.forceUpdate('errorDialog')
                                    if commandStr == 'finish':
                                        # update datasets
                                        tmpLog.info('updating datasets to finish')
                                        tmpStat = self.taskBufferIF.updateDatasetsToFinishTask_JEDI(jediTaskID, self.pid)
                                        if not tmpStat:
                                            tmpLog.info('wait until datasets are updated to finish')
                                        # ignore failGoalUnreached when manually finished
                                        tmpStat,taskSpec = self.taskBufferIF.getTaskWithID_JEDI(jediTaskID)
                                        tmpTaskSpec.splitRule = taskSpec.splitRule
                                        tmpTaskSpec.unsetFailGoalUnreached()
                                    if updateTaskStatus:
                                        tmpTaskSpec.status = JediTaskSpec.commandStatusMap()[commandStr]['done']
                                    tmpMsg = 'set task_status={0}'.format(tmpTaskSpec.status)
                                    tmpLog.sendMsg(tmpMsg,self.msgType)
                                    tmpLog.info(tmpMsg)
                                    tmpRet = self.taskBufferIF.updateTask_JEDI(tmpTaskSpec,{'jediTaskID':jediTaskID},
                                                                               setOldModTime=True)
                                    tmpLog.info('done with {0}'.format(str(tmpRet)))
                                    break
                                else:
                                    # kill only in the first loop
                                    if iLoop > 0:
                                        break
                                    # wait or kill jobs
                                    if commentStr and 'soft finish' in commentStr:
                                        queuedPandaIDs = self.taskBufferIF.getQueuedPandaIDsWithTask_JEDI(jediTaskID)
                                        tmpMsg = "trying to kill {0} queued jobs for soft finish".format(len(queuedPandaIDs))
                                        tmpLog.info(tmpMsg)
                                        tmpRet = self.taskBufferIF.killJobs(queuedPandaIDs,commentStr,'52',True)
                                        tmpMsg = "wating {0} jobs for soft finish".format(len(pandaIDs))
                                        tmpLog.info(tmpMsg)
                                        tmpRet = True
                                        tmpLog.info('done with {0}'.format(str(tmpRet)))
                                        break
                                    else:
                                        tmpMsg = "trying to kill {0} jobs".format(len(pandaIDs))
                                        tmpLog.info(tmpMsg)
                                        tmpLog.sendMsg(tmpMsg,self.msgType)
                                        if commandStr in ['finish']:
                                            # force kill
                                            tmpRet = self.taskBufferIF.killJobs(pandaIDs,commentStr,'52',True)
                                        elif commandStr in ['reassign']:
                                            # force kill
                                            tmpRet = self.taskBufferIF.killJobs(pandaIDs,commentStr,'51',True)
                                        else:
                                            # normal kill
                                            tmpRet = self.taskBufferIF.killJobs(pandaIDs,commentStr,'50',True)
                                        tmpLog.info('done with {0}'.format(str(tmpRet)))
                    elif commandStr in ['retry','incexec']:
                        tmpMsg = 'executing {0}'.format(commandStr)
                        tmpLog.info(tmpMsg)
                        tmpLog.sendMsg(tmpMsg,self.msgType)
                        # change task params for incexec
                        if commandStr == 'incexec':
                            try:
                                # read task params
                                taskParam = self.taskBufferIF.getTaskParamsWithID_JEDI(jediTaskID)
                                taskParamMap = RefinerUtils.decodeJSON(taskParam)
                                # remove some params
                                for newKey in ['nFiles','fixedSandbox']:
                                    try:
                                        del taskParamMap[newKey]
                                    except Exception:
                                        pass
                                # convert new params
                                newParamMap = RefinerUtils.decodeJSON(commentStr)
                                # change params
                                for newKey,newVal in iteritems(newParamMap):
                                    if newVal is None:
                                        # delete
                                        if newKey in taskParamMap:
                                            del taskParamMap[newKey]
                                    else:
                                        # change
                                        taskParamMap[newKey] = newVal
                                # overwrite sandbox
                                if 'fixedSandbox' in taskParamMap:
                                    # noBuild
                                    for tmpParam in taskParamMap['jobParameters']:
                                        if tmpParam['type'] == 'constant' and re.search('^-a [^ ]+$',tmpParam['value']) is not None:
                                            tmpParam['value'] = '-a {0}'.format(taskParamMap['fixedSandbox'])
                                    # build
                                    if 'buildSpec' in taskParamMap:
                                        taskParamMap['buildSpec']['archiveName'] = taskParamMap['fixedSandbox']
                                    # merge
                                    if 'mergeSpec' in taskParamMap:
                                        taskParamMap['mergeSpec']['jobParameters'] = \
                                            re.sub('-a [^ ]+','-a {0}'.format(taskParamMap['fixedSandbox']),taskParamMap['mergeSpec']['jobParameters'])
                                # encode new param
                                strTaskParams = RefinerUtils.encodeJSON(taskParamMap)
                                tmpRet = self.taskBufferIF.updateTaskParams_JEDI(jediTaskID,strTaskParams)
                                if tmpRet is not True:
                                    tmpLog.error('failed to update task params')
                                    continue
                            except Exception as e:
                                tmpLog.error('failed to change task params with {} {}'.format(str(e),
                                                                                              traceback.format_exc()))
                                continue
                        # retry child tasks
                        if 'sole ' in commentStr:
                            retryChildTasks = False
                        else:
                            retryChildTasks = True
                        # discard events
                        if 'discard ' in commentStr:
                            discardEvents = True
                        else:
                            discardEvents = False
                        # release un-staged files
                        if 'staged ' in commentStr:
                            releaseUnstaged = True
                        else:
                            releaseUnstaged = False
                        tmpRet,newTaskStatus = self.taskBufferIF.retryTask_JEDI(jediTaskID,commandStr,
                                                                                retryChildTasks=retryChildTasks,
                                                                                discardEvents=discardEvents,
                                                                                release_unstaged=releaseUnstaged)
                        if tmpRet is True:
                            tmpMsg = 'set task_status={0}'.format(newTaskStatus)
                            tmpLog.sendMsg(tmpMsg,self.msgType)
                            tmpLog.info(tmpMsg)
                        tmpLog.info('done with {0}'.format(tmpRet))
                    else:
                        tmpLog.error('unknown command')
            except Exception as e:
                errStr = '{} failed in runImpl() with {} {} '.format(self.__class__.__name__, str(e),
                                                                     traceback.format_exc())
                logger.error(errStr)



########## launch

def launcher(commuChannel,taskBufferIF,ddmIF,vos=None,prodSourceLabels=None):
    p = TaskCommando(commuChannel,taskBufferIF,ddmIF,vos,prodSourceLabels)
    p.start()



import re

import yaml

from pandajedi.jedimsgprocessor.base_msg_processor import BaseMsgProcPlugin

from pandacommon.pandalogger import logger_utils
from pandaserver.dataservice.DDMHandler import DDMHandler


# logger
base_logger = logger_utils.setup_logger(__name__.split('.')[-1])


# panda dataset callback message processing plugin
class PandaCallbackMsgProcPlugin(BaseMsgProcPlugin):

    def process(self, msg_obj):
        # logger
        tmp_log = logger_utils.make_logger(base_logger, method_name='process')
        # start
        # tmp_log.info('start')
        # tmp_log.debug('sub_id={0} ; msg_id={1}'.format(msg_obj.sub_id, msg_obj.msg_id))
        # parse yaml
        try:
            message_dict = yaml.safe_load(msg_obj.data)
        except Exception as e:
            err_str = 'failed to parse message yaml {2} , skipped. {0} : {1}'.format(e.__class__.__name__, e, msg_obj.data)
            tmp_log.error(err_str)
            raise
        # run
        try:
            to_continue = True
            dsn = 'UNKNOWN'
            # check event type
            event_type = message_dict['event_type']
            if event_type not in ['datasetlock_ok']:
                # tmp_log.debug('{0} skip'.format(event_type))
                to_continue = False
            if to_continue:
                # tmp_log.debug('{0} start'.format(event_type))
                message_payload = message_dict['payload']
                # only for _dis or _sub
                dsn = message_payload['name']
                if (re.search('_dis\d+$', dsn) is None) and (re.search('_sub\d+$', dsn) is None):
                    # tmp_log.debug('{0} is not _dis or _sub dataset, skip'.format(dsn))
                    to_continue = False
            if to_continue:
                tmp_log.debug('sub_id={0} ; msg_id={1}'.format(msg_obj.sub_id, msg_obj.msg_id))
                tmp_log.debug('{0} start'.format(event_type))
                # take action
                scope = message_payload['scope']
                site  = message_payload['rse']
                tmp_log.debug('{dsn} site={site} type={type}'.format(dsn=dsn, site=site, type=event_type))
                thr = DDMHandler(taskBuffer=self.tbIF, vuid=None, site=site, dataset=dsn, scope=scope)
                # just call run rather than start+join, to run it in main thread instead of spawning new thread
                thr.run()
                del thr
                tmp_log.debug('done {0}'.format(dsn))
        except Exception as e:
            err_str = 'failed to run, skipped. {0} : {1}'.format(e.__class__.__name__, e)
            tmp_log.error(err_str)
            raise
        # done
        # tmp_log.info('done')



import uuid
import time
import smtplib
import datetime

from pandajedi.jedicore import Interaction
from pandaserver.config import panda_config
from pandaserver.taskbuffer import EventServiceUtils

# port for SMTP server
smtpPortList = [25,587]


# wrapper to patch smtplib.stderr to send debug info to logger
class StderrLogger(object):
    def __init__(self,tmpLog):
        self.tmpLog = tmpLog
    def write(self,message):
        message = message.strip()
        if message != '':
            self.tmpLog.debug(message)


# wrapper of SMTP to redirect messages
class MySMTP (smtplib.SMTP):

    def set_log(self, tmp_log):
        self.tmpLog = tmp_log
        try:
            self.org_stderr = getattr(smtplib, 'stderr')
            setattr(smtplib, 'stderr', tmp_log)
        except Exception:
            self.org_stderr = None

    def _print_debug(self, *args):
        self.tmpLog.write(' '.join(map(str, args)))

    def reset_log(self):
        if self.org_stderr is not None:
            setattr(smtplib, 'stderr', self.org_stderr)


# base class for post process
class PostProcessorBase (object):

    # constructor
    def __init__(self,taskBufferIF,ddmIF):
        self.ddmIF = ddmIF
        self.taskBufferIF = taskBufferIF
        self.msgType = 'postprocessor'
        self.refresh()


    # refresh
    def refresh(self):
        self.siteMapper = self.taskBufferIF.getSiteMapper()



    # basic post procedure
    def doBasicPostProcess(self,taskSpec,tmpLog):
        # update task status
        taskSpec.lockedBy = None
        taskSpec.status = self.getFinalTaskStatus(taskSpec)
        if taskSpec.status == 'failed':
            # set dialog for preprocess
            if taskSpec.usePrePro() and not taskSpec.checkPreProcessed():
                taskSpec.setErrDiag('Preprocessing step failed',True)
        tmpMsg = 'set task_status={0}'.format(taskSpec.status)
        tmpLog.info(tmpMsg)
        tmpLog.sendMsg('set task_status={0}'.format(taskSpec.status),self.msgType)
        # update dataset
        for datasetSpec in taskSpec.datasetSpecList:
            if taskSpec.status in ['failed','broken','aborted']:
                datasetSpec.status = 'failed'
            else:
                # set dataset status
                if datasetSpec.type in ['output','log','lib']:
                    # normal output datasets
                    if datasetSpec.nFiles and datasetSpec.nFilesFinished and datasetSpec.nFiles > datasetSpec.nFilesFinished:
                        datasetSpec.status = 'finished'
                    else:
                        datasetSpec.status = 'done'
                elif datasetSpec.type.startswith('trn_') or datasetSpec.type.startswith('tmpl_'):
                    # set done for template or transient datasets
                    datasetSpec.status = 'done'
                else:
                    # not for input
                    continue
            # set nFiles
            if datasetSpec.type in ['output','log','lib']:
                datasetSpec.nFiles = datasetSpec.nFilesFinished
            self.taskBufferIF.updateDataset_JEDI(datasetSpec,{'datasetID':datasetSpec.datasetID,
                                                              'jediTaskID':datasetSpec.jediTaskID})
        # end time
        taskSpec.endTime = datetime.datetime.utcnow()
        # update task
        self.taskBufferIF.updateTask_JEDI(taskSpec,{'jediTaskID':taskSpec.jediTaskID},
                                          updateDEFT=True)
        # kill or kick child tasks
        if taskSpec.status in ['failed','broken','aborted']:
            self.taskBufferIF.killChildTasks_JEDI(taskSpec.jediTaskID,taskSpec.status)
        else:
            self.taskBufferIF.kickChildTasks_JEDI(taskSpec.jediTaskID)
        tmpLog.debug('doBasicPostProcess done with taskStatus={0}'.format(taskSpec.status))
        return



    # final procedure
    def doFinalProcedure(self,taskSpec,tmpLog):
        return self.SC_SUCCEEDED



    # send mail
    def sendMail(self,jediTaskID,fromAdd,toAdd,msgBody,nTry,fileBackUp,tmpLog):
        tmpLog.debug("sending notification to {0}\n{1}".format(toAdd,msgBody))
        for iTry in range(nTry):
            try:
                stderrLog = StderrLogger(tmpLog)
                smtpPort = smtpPortList[iTry % len(smtpPortList)]
                server = MySMTP(panda_config.emailSMTPsrv,smtpPort)
                server.set_debuglevel(1)
                server.set_log(stderrLog)
                server.ehlo()
                server.starttls()
                out = server.sendmail(fromAdd,toAdd,msgBody)
                tmpLog.debug(str(out))
                server.quit()
                break
            except Exception as e:
                if iTry+1 < nTry:
                    # sleep for retry
                    tmpLog.debug("sleep {0} due to {1}".format(iTry, str(e)))
                    time.sleep(30)
                else:
                    tmpLog.error("failed to send notification with {0}".format(str(e)))
                    if fileBackUp:
                        # write to file which is processed in add.py
                        mailFile = '{0}/jmail_{1}_{2}' % (panda_config.logdir, jediTaskID, uuid.uuid4())
                        oMail = open(mailFile,"w")
                        oMail.write(str(jediTaskID)+'\n'+toAdd+'\n'+msgBody)
                        oMail.close()
                break
        try:
            server.reset_log()
        except Exception:
            pass



    # return email sender
    def senderAddress(self):
        return panda_config.emailSender



    # get task completeness
    def getTaskCompleteness(self,taskSpec):
        nFiles = 0
        nFilesFinished = 0
        totalInputEvents = 0
        totalOkEvents = 0
        for datasetSpec in taskSpec.datasetSpecList:
            if datasetSpec.isMasterInput():
                nFiles += datasetSpec.nFiles
                nFilesFinished += datasetSpec.nFilesFinished
                try:
                    totalInputEvents += datasetSpec.nEvents
                except Exception:
                    pass
                try:
                    totalOkEvents += datasetSpec.nEventsUsed
                except Exception:
                    pass
        # completeness
        if totalInputEvents != 0:
            taskCompleteness = float(totalOkEvents)/float(totalInputEvents)*1000.0
        elif nFiles != 0:
            taskCompleteness = float(nFilesFinished)/float(nFiles)*1000.0
        else:
            taskCompleteness = 0
        return nFiles,nFilesFinished,totalInputEvents,totalOkEvents,taskCompleteness



    # get final task status
    def getFinalTaskStatus(self,taskSpec,checkParent=True,checkGoal=False):
        # count nFiles and nEvents
        nFiles,nFilesFinished,totalInputEvents,totalOkEvents,taskCompleteness = self.getTaskCompleteness(taskSpec)
        # set new task status
        if taskSpec.status == 'tobroken':
            status = 'broken'
        elif taskSpec.status == 'toabort':
            status = 'aborted'
        elif taskSpec.status == 'paused':
            status = 'paused'
        elif nFiles == nFilesFinished:
            # check parent status
            if checkParent and taskSpec.parent_tid not in [None,taskSpec.jediTaskID]:
                if self.taskBufferIF.getTaskStatus_JEDI(taskSpec.parent_tid) != 'done':
                    status = 'finished'
                else:
                    # check if input is mutable
                    inputMutable = False
                    for datasetSpec in taskSpec.datasetSpecList:
                        if datasetSpec.isMasterInput() and datasetSpec.state == 'mutable':
                            inputMutable = True
                            break
                    if inputMutable:
                        status = 'finished'
                    else:
                        status = 'done'
            else:
                status = 'done'
        elif nFilesFinished == 0:
            status = 'failed'
        else:
            status = 'finished'
        # task goal
        if taskSpec.goal is None:
            taskGoal = 1000
        else:
            taskGoal = taskSpec.goal
        # fail if goal is not reached
        if taskSpec.failGoalUnreached() and status == 'finished' and \
                (not taskSpec.useExhausted() or (taskSpec.useExhausted() and taskSpec.status in ['passed'])):
            if taskCompleteness < taskGoal:
                status = 'failed'
        # HPO tasks always go to finished
        if taskSpec.is_hpo_workflow():
            event_stat = self.taskBufferIF.get_event_statistics(taskSpec.jediTaskID)
            if event_stat is not None and event_stat.get(EventServiceUtils.ST_finished):
                status = 'finished'
        # check goal only
        if checkGoal:
            # no goal
            if taskSpec.goal is not None and taskCompleteness >= taskGoal:
                return True
            return False
        # return status
        return status



    # pre-check
    def doPreCheck(self,taskSpec,tmpLog):
        # send task to exhausted
        if taskSpec.useExhausted() and taskSpec.status not in ['passed'] \
                and self.getFinalTaskStatus(taskSpec) in ['finished'] \
                and not self.getFinalTaskStatus(taskSpec,checkGoal=True):
            taskSpec.status = 'exhausted'
            if self.getFinalTaskStatus(taskSpec, checkParent=False) == 'done':
                taskSpec.errorDialog = "exhausted since the parent task was incomplete"
            else:
                taskSpec.errorDialog = "exhausted since the task was incomplete"
            taskSpec.lockedBy = None
            taskSpec.lockedTime = None
            # update task
            tmpLog.info('set task_status={0}'.format(taskSpec.status))
            self.taskBufferIF.updateTask_JEDI(taskSpec,{'jediTaskID':taskSpec.jediTaskID},
                                              updateDEFT=True)
            # kick child tasks
            self.taskBufferIF.kickChildTasks_JEDI(taskSpec.jediTaskID)
            return True
        return False




Interaction.installSC(PostProcessorBase)



import sys
import time
import datetime
import traceback

from pandajedi.jedicore.ThreadUtils import ListWithLock,ThreadPool,WorkerThread
from pandajedi.jedicore import Interaction
from pandajedi.jedicore import JediException
from pandajedi.jedicore.MsgWrapper import MsgWrapper
from pandajedi.jedicore.FactoryBase import FactoryBase
from pandajedi.jedirefine import RefinerUtils
from .JediKnight import JediKnight

from pandajedi.jedicore.JediTaskSpec import JediTaskSpec
from pandajedi.jedicore.JediDatasetSpec import JediDatasetSpec
from pandajedi.jediconfig import jedi_config


# logger
from pandacommon.pandalogger.PandaLogger import PandaLogger
logger = PandaLogger().getLogger(__name__.split('.')[-1])


# worker class to refine TASK_PARAM to fill JEDI tables
class TaskRefiner (JediKnight,FactoryBase):

    # constructor
    def __init__(self, commuChannel, taskBufferIF, ddmIF, vos, prodSourceLabels):
        self.vos = self.parseInit(vos)
        self.prodSourceLabels = self.parseInit(prodSourceLabels)
        JediKnight.__init__(self,commuChannel,taskBufferIF,ddmIF,logger)
        FactoryBase.__init__(self, self.vos, self.prodSourceLabels, logger,
                             jedi_config.taskrefine.modConfig)


    # main
    def start(self):
        # start base classes
        JediKnight.start(self)
        FactoryBase.initializeMods(self, self.taskBufferIF, self.ddmIF)
        # go into main loop
        while True:
            startTime = datetime.datetime.utcnow()
            try:
                # get logger
                tmpLog = MsgWrapper(logger)
                tmpLog.debug('start')
                # loop over all vos
                for vo in self.vos:
                    # loop over all sourceLabels
                    for prodSourceLabel in self.prodSourceLabels:
                        # get the list of tasks to refine
                        tmpList = self.taskBufferIF.getTasksToRefine_JEDI(vo, prodSourceLabel)
                        if tmpList is None:
                            # failed
                            tmpLog.error('failed to get the list of tasks to refine')
                        else:
                            tmpLog.debug('got {0} tasks'.format(len(tmpList)))
                            # put to a locked list
                            taskList = ListWithLock(tmpList)
                            # make thread pool
                            threadPool = ThreadPool()
                            # get work queue mapper
                            workQueueMapper = self.taskBufferIF.getWorkQueueMap()
                            # make workers
                            nWorker = jedi_config.taskrefine.nWorkers
                            for iWorker in range(nWorker):
                                thr = TaskRefinerThread(taskList, threadPool,
                                                        self.taskBufferIF,
                                                        self.ddmIF,
                                                        self, workQueueMapper)
                                thr.start()
                            # join
                            threadPool.join()
            except Exception:
                errtype, errvalue = sys.exc_info()[:2]
                tmpLog.error('failed in {0}.start() with {1} {2}'.format(self.__class__.__name__,
                                                                         errtype.__name__, errvalue))
                tmpLog.error('Traceback: {0}'.format(traceback.format_exc()))
            # sleep if needed
            loopCycle = jedi_config.taskrefine.loopCycle
            timeDelta = datetime.datetime.utcnow() - startTime
            sleepPeriod = loopCycle - timeDelta.seconds
            if sleepPeriod > 0:
                time.sleep(sleepPeriod)
            # randomize cycle
            self.randomSleep(max_val=loopCycle)



# thread for real worker
class TaskRefinerThread (WorkerThread):

    # constructor
    def __init__(self, taskList, threadPool, taskbufferIF, ddmIF, implFactory, workQueueMapper):
        # initialize woker with no semaphore
        WorkerThread.__init__(self,None,threadPool,logger)
        # attributres
        self.taskList = taskList
        self.taskBufferIF = taskbufferIF
        self.ddmIF = ddmIF
        self.implFactory = implFactory
        self.workQueueMapper = workQueueMapper
        self.msgType = 'taskrefiner'



    # main
    def runImpl(self):
        while True:
            try:
                # get a part of list
                nTasks = 10
                taskList = self.taskList.get(nTasks)
                # no more datasets
                if len(taskList) == 0:
                    self.logger.info('{0} terminating since no more items'.format(self.__class__.__name__))
                    return
                # loop over all tasks
                for jediTaskID,splitRule,taskStatus,parent_tid in taskList:
                    # make logger
                    tmpLog = MsgWrapper(self.logger,'< jediTaskID={0} >'.format(jediTaskID))
                    tmpLog.debug('start')
                    tmpStat = Interaction.SC_SUCCEEDED
                    errStr = ''
                    # read task parameters
                    try:
                        taskParam = None
                        taskParam = self.taskBufferIF.getTaskParamsWithID_JEDI(jediTaskID)
                        taskParamMap = RefinerUtils.decodeJSON(taskParam)
                    except Exception:
                        errtype,errvalue = sys.exc_info()[:2]
                        errStr = 'conversion to map from json failed with {0}:{1}'.format(errtype.__name__,errvalue)
                        tmpLog.debug(taskParam)
                        tmpLog.error(errStr)
                        continue
                        tmpStat = Interaction.SC_FAILED
                    # get impl
                    if tmpStat == Interaction.SC_SUCCEEDED:
                        tmpLog.info('getting Impl')
                        try:
                            # get VO and sourceLabel
                            vo = taskParamMap['vo']
                            prodSourceLabel = taskParamMap['prodSourceLabel']
                            taskType = taskParamMap['taskType']
                            tmpLog.info('vo={0} sourceLabel={1} taskType={2}'.format(vo,prodSourceLabel,taskType))
                            # get impl
                            impl = self.implFactory.instantiateImpl(vo, prodSourceLabel, taskType,
                                                                    self.taskBufferIF, self.ddmIF)
                            if impl is None:
                                # task refiner is undefined
                                errStr = 'task refiner is undefined for vo={0} sourceLabel={1}'.format(vo,prodSourceLabel)
                                tmpLog.error(errStr)
                                tmpStat = Interaction.SC_FAILED
                        except Exception:
                            errtype,errvalue = sys.exc_info()[:2]
                            errStr = 'failed to get task refiner with {0}:{1}'.format(errtype.__name__,errvalue)
                            tmpLog.error(errStr)
                            tmpStat = Interaction.SC_FAILED
                    # extract common parameters
                    if tmpStat == Interaction.SC_SUCCEEDED:
                        tmpLog.info('extracting common')
                        try:
                            # initalize impl
                            impl.initializeRefiner(tmpLog)
                            impl.oldTaskStatus = taskStatus
                            # extract common parameters
                            impl.extractCommon(jediTaskID, taskParamMap, self.workQueueMapper, splitRule)
                            # set parent tid
                            if parent_tid not in [None,jediTaskID]:
                                impl.taskSpec.parent_tid = parent_tid
                        except Exception:
                            errtype, errvalue = sys.exc_info()[:2]
                            # on hold in case of external error
                            if errtype == JediException.ExternalTempError:
                                tmpErrStr = 'pending due to external problem. {0}'.format(errvalue)
                                setFrozenTime = True
                                impl.taskSpec.status = taskStatus
                                impl.taskSpec.setOnHold()
                                impl.taskSpec.setErrDiag(tmpErrStr)
                                # not to update some task attributes
                                impl.taskSpec.resetRefinedAttrs()
                                tmpLog.info(tmpErrStr)
                                self.taskBufferIF.updateTask_JEDI(impl.taskSpec,{'jediTaskID':impl.taskSpec.jediTaskID},
                                                                  oldStatus=[taskStatus],
                                                                  insertUnknown=impl.unknownDatasetList,
                                                                  setFrozenTime=setFrozenTime)
                                continue
                            errStr = 'failed to extract common parameters with {0}:{1} {2}'.format(errtype.__name__,errvalue,
                                                                                                   traceback.format_exc())
                            tmpLog.error(errStr)
                            tmpStat = Interaction.SC_FAILED
                    # check attribute length
                    if tmpStat == Interaction.SC_SUCCEEDED:
                        tmpLog.info('checking attribute length')
                        if not impl.taskSpec.checkAttrLength():
                            tmpLog.error(impl.taskSpec.errorDialog)
                            tmpStat = Interaction.SC_FAILED
                    # staging
                    if tmpStat == Interaction.SC_SUCCEEDED:
                        if 'toStaging' in taskParamMap and taskStatus not in ['staged', 'rerefine']:
                            errStr = 'wait until staging is done'
                            impl.taskSpec.status = 'staging'
                            impl.taskSpec.oldStatus = taskStatus
                            impl.taskSpec.setErrDiag(errStr)
                            # not to update some task attributes
                            impl.taskSpec.resetRefinedAttrs()
                            tmpLog.info(errStr)
                            self.taskBufferIF.updateTask_JEDI(impl.taskSpec, {'jediTaskID':impl.taskSpec.jediTaskID},
                                                              oldStatus=[taskStatus], updateDEFT=False, setFrozenTime=False)
                            continue
                    # check parent
                    noWaitParent = False
                    parentState = None
                    if tmpStat == Interaction.SC_SUCCEEDED:
                        if parent_tid not in [None,jediTaskID]:
                            tmpLog.info('check parent task')
                            try:
                                tmpStat = self.taskBufferIF.checkParentTask_JEDI(parent_tid)
                                parentState = tmpStat
                                if tmpStat == 'completed':
                                    # parent is done
                                    tmpStat = Interaction.SC_SUCCEEDED
                                elif tmpStat is None or tmpStat == 'running':
                                    if not impl.taskSpec.noWaitParent():
                                        # parent is running
                                        errStr = 'pending until parent task {0} is done'.format(parent_tid)
                                        impl.taskSpec.status = taskStatus
                                        impl.taskSpec.setOnHold()
                                        impl.taskSpec.setErrDiag(errStr)
                                        # not to update some task attributes
                                        impl.taskSpec.resetRefinedAttrs()
                                        tmpLog.info(errStr)
                                        self.taskBufferIF.updateTask_JEDI(impl.taskSpec,{'jediTaskID':impl.taskSpec.jediTaskID},
                                                                          oldStatus=[taskStatus],setFrozenTime=False)
                                        continue
                                    else:
                                        # not wait for parent
                                        tmpStat = Interaction.SC_SUCCEEDED
                                        noWaitParent = True
                                else:
                                    # parent is corrupted
                                    tmpStat = Interaction.SC_FAILED
                                    tmpErrStr = 'parent task {0} failed to complete'.format(parent_tid)
                                    impl.taskSpec.setErrDiag(tmpErrStr)
                            except Exception:
                                errtype,errvalue = sys.exc_info()[:2]
                                errStr = 'failed to check parent task with {0}:{1}'.format(errtype.__name__,errvalue)
                                tmpLog.error(errStr)
                                tmpStat = Interaction.SC_FAILED
                    # refine
                    if tmpStat == Interaction.SC_SUCCEEDED:
                        tmpLog.info('refining with {0}'.format(impl.__class__.__name__))
                        try:
                            tmpStat = impl.doRefine(jediTaskID,taskParamMap)
                        except Exception:
                            errtype,errvalue = sys.exc_info()[:2]
                            # wait unknown input if noWaitParent or waitInput
                            if ((impl.taskSpec.noWaitParent() or impl.taskSpec.waitInput()) \
                                    and errtype == JediException.UnknownDatasetError) or parentState == 'running' \
                                    or errtype == Interaction.JEDITemporaryError:
                                if impl.taskSpec.noWaitParent() or parentState == 'running':
                                    tmpErrStr = 'pending until parent produces input'
                                    setFrozenTime=False
                                elif errtype == Interaction.JEDITemporaryError:
                                    tmpErrStr = 'pending due to DDM problem. {0}'.format(errvalue)
                                    setFrozenTime=True
                                else:
                                    tmpErrStr = 'pending until input is staged'
                                    setFrozenTime=True
                                impl.taskSpec.status = taskStatus
                                impl.taskSpec.setOnHold()
                                impl.taskSpec.setErrDiag(tmpErrStr)
                                # not to update some task attributes
                                impl.taskSpec.resetRefinedAttrs()
                                tmpLog.info(tmpErrStr)
                                self.taskBufferIF.updateTask_JEDI(impl.taskSpec,{'jediTaskID':impl.taskSpec.jediTaskID},
                                                                  oldStatus=[taskStatus],
                                                                  insertUnknown=impl.unknownDatasetList,
                                                                  setFrozenTime=setFrozenTime)
                                continue
                            else:
                                errStr  = 'failed to refine task with {0}:{1}'.format(errtype.__name__,errvalue)
                                tmpLog.error(errStr)
                                tmpStat = Interaction.SC_FAILED
                    # register
                    if tmpStat != Interaction.SC_SUCCEEDED:
                        tmpLog.error('failed to refine the task')
                        if impl is None or impl.taskSpec is None:
                            tmpTaskSpec = JediTaskSpec()
                            tmpTaskSpec.jediTaskID = jediTaskID
                        else:
                            tmpTaskSpec = impl.taskSpec
                        tmpTaskSpec.status = 'tobroken'
                        if errStr != '':
                            tmpTaskSpec.setErrDiag(errStr,True)
                        self.taskBufferIF.updateTask_JEDI(tmpTaskSpec,{'jediTaskID':tmpTaskSpec.jediTaskID},oldStatus=[taskStatus])
                    else:
                        tmpLog.info('registering')
                        # fill JEDI tables
                        try:
                            # enable protection against task duplication
                            if 'uniqueTaskName' in taskParamMap and taskParamMap['uniqueTaskName'] and \
                                    not impl.taskSpec.checkPreProcessed():
                                uniqueTaskName = True
                            else:
                                uniqueTaskName = False
                            strTaskParams = None
                            if impl.updatedTaskParams is not None:
                                strTaskParams = RefinerUtils.encodeJSON(impl.updatedTaskParams)
                            if taskStatus in ['registered', 'staged']:
                                # unset pre-process flag
                                if impl.taskSpec.checkPreProcessed():
                                    impl.taskSpec.setPostPreProcess()
                                # full registration
                                tmpStat,newTaskStatus = self.taskBufferIF.registerTaskInOneShot_JEDI(jediTaskID,impl.taskSpec,
                                                                                                     impl.inMasterDatasetSpec,
                                                                                                     impl.inSecDatasetSpecList,
                                                                                                     impl.outDatasetSpecList,
                                                                                                     impl.outputTemplateMap,
                                                                                                     impl.jobParamsTemplate,
                                                                                                     strTaskParams,
                                                                                                     impl.unmergeMasterDatasetSpec,
                                                                                                     impl.unmergeDatasetSpecMap,
                                                                                                     uniqueTaskName,
                                                                                                     taskStatus)
                                if not tmpStat:
                                    tmpErrStr = 'failed to register the task to JEDI in a single shot'
                                    tmpLog.error(tmpErrStr)
                                    tmpTaskSpec = JediTaskSpec()
                                    tmpTaskSpec.status = newTaskStatus
                                    tmpTaskSpec.errorDialog = impl.taskSpec.errorDialog
                                    tmpTaskSpec.setErrDiag(tmpErrStr, True)
                                    self.taskBufferIF.updateTask_JEDI(tmpTaskSpec,
                                                                      {'jediTaskID': impl.taskSpec.jediTaskID},
                                                                      oldStatus=[taskStatus])
                                tmpMsg = 'set task_status={0}'.format(newTaskStatus)
                                tmpLog.info(tmpMsg)
                                tmpLog.sendMsg(tmpMsg,self.msgType)
                            else:
                                # disable scouts if previous attempt didn't use it
                                if not impl.taskSpec.useScout(splitRule):
                                    impl.taskSpec.setUseScout(False)
                                # disallow to reset some attributes
                                for attName in ['ramCount', 'walltime', 'cpuTime', 'startTime', 'cpuTimeUnit',
                                                'outDiskCount', 'workDiskCount', 'ioIntensity', 'diskIO']:
                                    impl.taskSpec.resetChangedAttr(attName)
                                # update task with new params
                                self.taskBufferIF.updateTask_JEDI(impl.taskSpec,{'jediTaskID':impl.taskSpec.jediTaskID},
                                                                  oldStatus=[taskStatus])
                                # appending for incremetnal execution
                                tmpStat = self.taskBufferIF.appendDatasets_JEDI(jediTaskID,impl.inMasterDatasetSpec,
                                                                                impl.inSecDatasetSpecList)
                                if not tmpStat:
                                    tmpLog.error('failed to append datasets for incexec')
                        except Exception:
                            errtype,errvalue = sys.exc_info()[:2]
                            tmpErrStr = 'failed to register the task to JEDI with {0}:{1}'.format(errtype.__name__,errvalue)
                            tmpLog.error(tmpErrStr)
                        else:
                            tmpLog.info('done')
            except Exception:
                errtype,errvalue = sys.exc_info()[:2]
                logger.error('{0} failed in runImpl() with {1}:{2}'.format(self.__class__.__name__,errtype.__name__,errvalue))



########## lauch

def launcher(commuChannel,taskBufferIF,ddmIF,vos=None,prodSourceLabels=None):
    p = TaskRefiner(commuChannel,taskBufferIF,ddmIF,vos,prodSourceLabels)
    p.start()



from userinterface import Client

import sys
import uuid

site = sys.argv[1]

dsName = 'mc12_8TeV.189659.gg2VVPythia8_AU2CT10_ggH125p5_VV_2mu2numu_m2l4_2pt3.evgen.EVNT.e2872_tid01461041_00'
logDatasetName = 'panda.jeditest.log.{0}'.format(uuid.uuid4())

taskParamMap = {}

#taskParamMap['nFilesPerJob'] = 1
taskParamMap['nFiles'] = 1
taskParamMap['nEventsPerFile']   = 5000
taskParamMap['nEventsPerWorker'] = 10 #5000
taskParamMap['nEventsPerJob'] = 1000
taskParamMap['nEvents'] = 1000
#taskParamMap['skipScout'] = True
taskParamMap['objectStore'] = 'root^atlas-objectstore.cern.ch^/atlas/eventservice'
taskParamMap['taskName'] = str(uuid.uuid4())
taskParamMap['userName'] = 'pandasrv1'
taskParamMap['vo'] = 'atlas'
taskParamMap['taskPriority'] = 900
#taskParamMap['reqID'] = reqIdx
taskParamMap['architecture'] = 'x86_64-slc5-gcc43-opt'
taskParamMap['transUses'] = 'Atlas-17.2.11'
taskParamMap['transHome'] = 'AtlasProduction-17.2.11.8'
taskParamMap['transPath'] = 'AtlasG4_tf.py'
taskParamMap['processingType'] = 'simul'
taskParamMap['prodSourceLabel'] = 'ptest'
taskParamMap['taskType'] = 'prod'
taskParamMap['workingGroup'] = 'AP_Higgs'
taskParamMap['coreCount'] = 1
taskParamMap['walltime'] = 1
taskParamMap['cloud'] = 'US'
taskParamMap['site'] = site
taskParamMap['log'] = {'dataset': logDatasetName,
                       'type':'template',
                       'param_type':'log',
                       'token':'ATLASDATADISK',
                       'offset':1000,
                       'value':'{0}.${{SN}}.log.tgz'.format(logDatasetName)}
outDatasetName = 'panda.jeditest.HITS.{0}'.format(uuid.uuid4())


taskParamMap['jobParameters'] = [
    {'type':'constant',
     'value': '--AMITag s1831 --DBRelease=current --athenaopts=--preloadlib=${ATLASMKLLIBDIR_PRELOAD}/libimf.so --conditionsTag=OFLCOND-MC12-SIM-00',
     },
    {'type':'template',
     'value':'firstEvent=${FIRSTEVENT}',
     'param_type':'number',
     },
    {'type':'constant',
     'value': '--geometryVersion=ATLAS-GEO-21-02-02_VALIDATION',
     },
    {'type':'template',
     'param_type':'input',
     'value':'--inputEvgenFile=${IN}',
     'dataset':dsName,
     },
    {'type':'template',
     'value':'--maxEvents=${MAXEVENTS}',
     'param_type':'number',
     },
    {'type':'template',
     'param_type':'output',
     'token':'ATLASDATADISK',
     'value':' --outputHitsFile={0}.${{SN}}.pool.root'.format(outDatasetName),
     'dataset':outDatasetName,
     },
    {'type':'constant',
     'value':'--physicsList=QGSP_BERT --postInclude=RecJobTransforms/UseFrontierFallbackDBRelease.py --preInclude=SimulationJobOptions/preInclude.FrozenShowersFCalOnly.py,SimulationJobOptions/preInclude.BeamPipeKill.py',
     },
    {'type':'template',
     'value':'--skipEvents=${SKIPEVENTS}',
     'param_type':'number',
     },
    {'type':'template',
     'value':'--randomSeed=${RNDMSEED}',
     'param_type':'number',
     },
    ]

taskParamMap['esmergeSpec'] = {}
taskParamMap['esmergeSpec']['transPath'] = 'Merge_trf.py'
taskParamMap['esmergeSpec']['jobParameters'] = "aaa bbb"

print(Client.insertTaskParams(taskParamMap))



from pandajedi.jedicore.FactoryBase import FactoryBase
from pandajedi.jediconfig import jedi_config

# logger
from pandacommon.pandalogger.PandaLogger import PandaLogger
logger = PandaLogger().getLogger(__name__.split('.')[-1])


# factory class for throttling
class JobThrottler (FactoryBase):

    # constructor
    def __init__(self,vo,sourceLabel):
        FactoryBase.__init__(self, vo, sourceLabel, logger,
                             jedi_config.jobthrottle.modConfig)


    # main
    def toBeThrottled(self, vo, sourceLabel, cloudName, workQueue, resourceType):
        impl = self.getImpl(vo, sourceLabel)
        retVal = impl.toBeThrottled(vo, sourceLabel, cloudName, workQueue, resourceType)
        # retrieve min priority and max number of jobs from concrete class
        self.minPriority = impl.minPriority
        self.maxNumJobs = impl.maxNumJobs
        self.lackOfJobs = impl.underNqLimit
        return retVal



    # check throttle level
    def mergeThrottled(self, vo, sourceLabel, thrLevel):
        impl = self.getImpl(vo, sourceLabel)
        return impl.mergeThrottled(thrLevel)



import sys
import time
import datetime

from pandajedi.jedicore.ThreadUtils import ListWithLock,ThreadPool,WorkerThread
from pandajedi.jedicore import Interaction
from pandajedi.jedicore.MsgWrapper import MsgWrapper
from pandajedi.jedicore.FactoryBase import FactoryBase
from .JediKnight import JediKnight

from pandajedi.jediconfig import jedi_config


# logger
from pandacommon.pandalogger.PandaLogger import PandaLogger
logger = PandaLogger().getLogger(__name__.split('.')[-1])


# worker class to refine TASK_PARAM to fill JEDI tables 
class TaskBroker (JediKnight,FactoryBase):

    # constructor
    def __init__(self,commuChannel,taskBufferIF,ddmIF,vos,prodSourceLabels):
        self.vos = self.parseInit(vos)
        self.prodSourceLabels = self.parseInit(prodSourceLabels)
        JediKnight.__init__(self,commuChannel,taskBufferIF,ddmIF,logger)
        FactoryBase.__init__(self,self.vos,self.prodSourceLabels,logger,
                             jedi_config.taskbroker.modConfig)


    # main
    def start(self):
        # start base classes
        JediKnight.start(self)
        FactoryBase.initializeMods(self,self.taskBufferIF,self.ddmIF)
        # go into main loop
        while True:
            startTime = datetime.datetime.utcnow()
            try:
                # get logger
                tmpLog = MsgWrapper(logger)
                tmpLog.debug('start TaskBroker')
                # get work queue mapper
                workQueueMapper = self.taskBufferIF.getWorkQueueMap()
                resource_types = self.taskBufferIF.load_resource_types()

                # loop over all vos
                for vo in self.vos:
                    # loop over all sourceLabels
                    for prodSourceLabel in self.prodSourceLabels:
                        # loop over all work queues
                        for workQueue in workQueueMapper.getAlignedQueueList(vo, prodSourceLabel):
                            for resource_type in resource_types:
                                wq_name = '_'.join(workQueue.queue_name.split(' '))
                                msgLabel = 'vo={0} label={1} queue={2} resource_type={3}: '.\
                                    format(vo, prodSourceLabel, wq_name, resource_type.resource_name)
                                tmpLog.debug(msgLabel+'start')
                                # get the list of tasks to check
                                tmpList = self.taskBufferIF.getTasksToCheckAssignment_JEDI(vo, prodSourceLabel,
                                                                                           workQueue,
                                                                                           resource_type.resource_name)
                                if tmpList is None:
                                    # failed
                                    tmpLog.error(msgLabel+'failed to get the list of tasks to check')
                                else:
                                    tmpLog.debug(msgLabel+'got tasks_to_check={0}'.format(len(tmpList)))
                                    # put to a locked list
                                    taskList = ListWithLock(tmpList)
                                    # make thread pool
                                    threadPool = ThreadPool()
                                    # make workers
                                    nWorker = jedi_config.taskbroker.nWorkers
                                    for iWorker in range(nWorker):
                                        thr = TaskCheckerThread(taskList,threadPool,
                                                                self.taskBufferIF,
                                                                self.ddmIF,self,
                                                                vo,prodSourceLabel)
                                        thr.start()
                                    # join
                                    threadPool.join()
                                # get the list of tasks to assign
                                tmpList = self.taskBufferIF.getTasksToAssign_JEDI(vo, prodSourceLabel,
                                                                                  workQueue, resource_type.resource_name)
                                if tmpList is None:
                                    # failed
                                    tmpLog.error(msgLabel+'failed to get the list of tasks to assign')
                                else:
                                    tmpLog.debug(msgLabel+'got tasks_to_assign={0}'.format(len(tmpList)))
                                    # put to a locked list
                                    taskList = ListWithLock(tmpList)
                                    # make thread pool
                                    threadPool = ThreadPool()
                                    # make workers
                                    nWorker = jedi_config.taskbroker.nWorkers
                                    for iWorker in range(nWorker):
                                        thr = TaskBrokerThread(taskList,threadPool,
                                                               self.taskBufferIF,
                                                               self.ddmIF,self,
                                                               vo,prodSourceLabel,
                                                               workQueue, resource_type.resource_name)
                                        thr.start()
                                    # join
                                    threadPool.join()
                                tmpLog.debug(msgLabel+'done')
            except Exception:
                errtype,errvalue = sys.exc_info()[:2]
                tmpLog.error('failed in {0}.start() with {1} {2}'.format(self.__class__.__name__,
                                                                         errtype.__name__,errvalue))
            tmpLog.debug('done')                                
            # sleep if needed
            loopCycle = jedi_config.taskbroker.loopCycle
            timeDelta = datetime.datetime.utcnow() - startTime
            sleepPeriod = loopCycle - timeDelta.seconds
            if sleepPeriod > 0:
                time.sleep(sleepPeriod)
            # randomize cycle
            self.randomSleep(max_val=loopCycle)



# thread for real worker
class TaskCheckerThread (WorkerThread):

    # constructor
    def __init__(self,taskList,threadPool,taskbufferIF,ddmIF,implFactory,
                 vo,prodSourceLabel):
        # initialize woker with no semaphore
        WorkerThread.__init__(self,None,threadPool,logger)
        # attributres
        self.taskList = taskList
        self.taskBufferIF = taskbufferIF
        self.ddmIF = ddmIF.getInterface(vo)
        self.implFactory = implFactory
        self.vo = vo
        self.prodSourceLabel = prodSourceLabel


    # main
    def runImpl(self):
        while True:
            try:
                # get a part of list
                nTasks = 100
                taskList = self.taskList.get(nTasks)
                totalTasks,idxTasks = self.taskList.stat()
                # no more datasets
                if len(taskList) == 0:
                    self.logger.debug('{0} terminating since no more items'.format(self.__class__.__name__))
                    return
                # make logger
                tmpLog = MsgWrapper(self.logger)
                tmpLog.info('start TaskCheckerThread {0}/{1} for jediTaskID={2}'.format(idxTasks,totalTasks,taskList))
                tmpStat = Interaction.SC_SUCCEEDED
                # get TaskSpecs
                taskSpecList = []
                for jediTaskID in taskList:
                    tmpRet,taskSpec = self.taskBufferIF.getTaskWithID_JEDI(jediTaskID,False)
                    if tmpRet and taskSpec is not None:
                        taskSpecList.append(taskSpec)
                    else:
                        tmpLog.error('failed to get taskSpec for jediTaskID={0}'.format(jediTaskID))
                if taskSpecList != []:
                    # get impl                    
                    if tmpStat == Interaction.SC_SUCCEEDED:                    
                        tmpLog.info('getting Impl')
                        try:
                            impl = self.implFactory.getImpl(self.vo,self.prodSourceLabel)
                            if impl is None:
                                # task brokerage is undefined
                                tmpLog.error('task broker is undefined for vo={0} sourceLabel={1}'.format(self.vo,self.prodSourceLabel))
                                tmpStat = Interaction.SC_FAILED
                        except Exception:
                            errtype,errvalue = sys.exc_info()[:2]
                            tmpLog.error('getImpl failed with {0}:{1}'.format(errtype.__name__,errvalue))
                            tmpStat = Interaction.SC_FAILED
                    # check
                    if tmpStat == Interaction.SC_SUCCEEDED:
                        tmpLog.info('check with {0}'.format(impl.__class__.__name__))
                        try:
                            tmpStat,taskCloudMap = impl.doCheck(taskSpecList)
                        except Exception:
                            errtype,errvalue = sys.exc_info()[:2]
                            tmpLog.error('doCheck failed with {0}:{1}'.format(errtype.__name__,errvalue))
                            tmpStat = Interaction.SC_FAILED
                    # update
                    if tmpStat != Interaction.SC_SUCCEEDED:
                        tmpLog.error('failed to check assignment')
                    else:
                        tmpRet = self.taskBufferIF.setCloudToTasks_JEDI(taskCloudMap)
                        tmpLog.info('done with {0} for {1}'.format(tmpRet,str(taskCloudMap)))
            except Exception:
                errtype,errvalue = sys.exc_info()[:2]
                logger.error('{0} failed in runImpl() with {1}:{2}'.format(self.__class__.__name__,errtype.__name__,errvalue))



# thread for real worker
class TaskBrokerThread (WorkerThread):

    # constructor
    def __init__(self, taskList, threadPool, taskbufferIF, ddmIF, implFactory,
                 vo, prodSourceLabel, workQueue, resource_name):
        # initialize woker with no semaphore
        WorkerThread.__init__(self,None,threadPool,logger)
        # attributres
        self.taskList = taskList
        self.taskBufferIF = taskbufferIF
        self.ddmIF = ddmIF.getInterface(vo)
        self.implFactory = implFactory
        self.vo = vo
        self.prodSourceLabel = prodSourceLabel
        self.workQueue = workQueue
        self.resource_name = resource_name

    # main
    def runImpl(self):
        while True:
            try:
                # get a part of list
                nTasks = 100
                taskList = self.taskList.get(nTasks)
                totalTasks,idxTasks = self.taskList.stat()
                # no more datasets
                if len(taskList) == 0:
                    self.logger.debug('{0} terminating since no more items'.format(self.__class__.__name__))
                    return
                # make logger
                tmpLog = MsgWrapper(self.logger)
                tmpLog.info('start TaskBrokerThread {0}/{1} for jediTaskID={2}'.format(idxTasks,totalTasks,taskList))
                tmpStat = Interaction.SC_SUCCEEDED
                # get TaskSpecs
                tmpListToAssign = []
                for tmpTaskItem in taskList:
                    tmpListItem = self.taskBufferIF.getTasksToBeProcessed_JEDI(None, None, None, None, None,
                                                                               simTasks=[tmpTaskItem],
                                                                               readMinFiles=True)
                    if tmpListItem is None:
                        # failed
                        tmpLog.error('failed to get the input chunks for jediTaskID={0}'.format(tmpTaskItem))
                        tmpStat = Interaction.SC_FAILED
                        break
                    tmpListToAssign += tmpListItem
                # get impl                    
                if tmpStat == Interaction.SC_SUCCEEDED:                    
                    tmpLog.info('getting Impl')
                    try:
                        impl = self.implFactory.getImpl(self.vo,self.prodSourceLabel)
                        if impl is None:
                            # task refiner is undefined
                            tmpLog.error('task broker is undefined for vo={0} sourceLabel={1}'.format(self.vo,self.prodSourceLabel))
                            tmpStat = Interaction.SC_FAILED
                    except Exception:
                        errtype,errvalue = sys.exc_info()[:2]
                        tmpLog.error('getImpl failed with {0}:{1}'.format(errtype.__name__,errvalue))
                        tmpStat = Interaction.SC_FAILED
                # brokerage
                if tmpStat == Interaction.SC_SUCCEEDED:
                    tmpLog.info('brokerage with {0} for {1} tasks '.format(impl.__class__.__name__,len(tmpListToAssign)))
                    try:
                        tmpStat = impl.doBrokerage(tmpListToAssign, self.vo,
                                                   self.prodSourceLabel, self.workQueue, self.resource_name)
                    except Exception:
                        errtype,errvalue = sys.exc_info()[:2]
                        tmpLog.error('doBrokerage failed with {0}:{1}'.format(errtype.__name__,errvalue))
                        tmpStat = Interaction.SC_FAILED
                # register
                if tmpStat != Interaction.SC_SUCCEEDED:
                    tmpLog.error('failed')
                else:
                    tmpLog.info('done')                    
            except Exception:
                errtype,errvalue = sys.exc_info()[:2]
                logger.error('{0} failed in runImpl() with {1}:{2}'.format(self.__class__.__name__,errtype.__name__,errvalue))
        


########## launch
                
def launcher(commuChannel,taskBufferIF,ddmIF,vos=None,prodSourceLabels=None):
    p = TaskBroker(commuChannel,taskBufferIF,ddmIF,vos,prodSourceLabels)
    p.start()



from pandajedi.jediconfig import jedi_config

from pandajedi.jedicore import Interaction

# interface to JediTaskBuffer
class JediTaskBufferInterface:

    # constructor
    def __init__(self):
        self.interface = None


    # setup interface
    def setupInterface(self):
        vo = 'any'
        maxSize = jedi_config.db.nWorkers
        moduleName = 'pandajedi.jedicore.JediTaskBuffer'
        className  = 'JediTaskBuffer'
        self.interface = Interaction.CommandSendInterface(vo,maxSize,
                                                          moduleName,
                                                          className)
        self.interface.initialize()


    # method emulation
    def __getattr__(self,attrName):
        return getattr(self.interface,attrName)


if __name__ == '__main__':
    def dummyClient(dif,stime):
        print("client test")
        import time
        for i in range(3):
            #time.sleep(i*stime)
            try:
                print(dif.getCloudList())
            except Exception:
                print("exp")
        print('client done')

    dif = JediTaskBufferInterface()
    dif.setupInterface()
    print("master test")
    print(dif.getCloudList())
    print("master done")
    import multiprocessing
    pList = []
    for i in range(5):
        p = multiprocessing.Process(target=dummyClient,
                                    args=(dif,i))
        pList.append(p)
        p.start()



import copy
import math
import random

from six import iteritems

try:
    long()
except Exception:
    long = int

from .JediTaskSpec import JediTaskSpec
from . import JediCoreUtils

from pandacommon.pandalogger.PandaLogger import PandaLogger
logger = PandaLogger().getLogger(__name__.split('.')[-1])

# class for input
class InputChunk:

    # default output size 2G + 500MB (safety merging)
    defaultOutputSize = 2500 * 1024 * 1024

    def __str__(self):
        sb = []
        for key in self.__dict__:
            sb.append("{key}='{value}'".format(key=key, value=self.__dict__[key]))

        return ', '.join(sb)

    def __repr__(self):
        return self.__str__()

    # constructor
    def __init__(self,taskSpec,masterDataset=None,secondaryDatasetList=[], ramCount=0):
        # task spec
        self.taskSpec = taskSpec
        # the list of secondary datasets
        if secondaryDatasetList is None:
            self.secondaryDatasetList = []
        else:
            self.secondaryDatasetList = secondaryDatasetList
        # the list of site candidates
        self.siteCandidates = {}
        # the list of site candidates for jumbo jobs
        self.siteCandidatesJumbo = {}
        # the name of master index
        self.masterIndexName = None
        # dataset mapping including indexes of files/events
        self.datasetMap = {}
        # the master dataset
        self.masterDataset = None
        self.addMasterDS(masterDataset)
        # the list of secondary datasets
        self.secondaryDatasetList = []
        for secondaryDS in secondaryDatasetList:
            self.addSecondaryDS(secondaryDS)
        # read in a block
        self.readBlock = None
        # merging
        self.isMerging = False
        # use scout
        self.useScoutFlag = None
        # memory requirements for the inputChunk
        self.ramCount = ramCount
        # flag to set if inputchunk is empty
        self.isEmpty = False
        # flag to use jumbo jobs
        self.useJumbo = None
        # checkpoint of used counters
        self.file_checkpoints = {}
        # list of bootstrapped sites
        self.bootstrapped = set()



    # add master dataset
    def addMasterDS(self,masterDataset):
        if masterDataset is not None:
            self.masterDataset = masterDataset
            self.masterIndexName = self.masterDataset.datasetID
            self.datasetMap[self.masterDataset.datasetID] = {'used':0,'datasetSpec':masterDataset}



    # add secondary dataset
    def addSecondaryDS(self,secondaryDataset):
        if secondaryDataset not in self.secondaryDatasetList:
            self.secondaryDatasetList.append(secondaryDataset)
            self.datasetMap[secondaryDataset.datasetID] = {'used':0,'datasetSpec':secondaryDataset}



    # return list of datasets
    def getDatasets(self,includePseudo=False):
        dataList = []
        if self.masterDataset is not None:
            dataList.append(self.masterDataset)
        dataList += self.secondaryDatasetList
        # ignore pseudo datasets
        if not includePseudo:
            newDataList = []
            for datasetSpec in dataList:
                if not datasetSpec.isPseudo():
                    newDataList.append(datasetSpec)
            dataList = newDataList
        return dataList



    # return dataset with datasetID
    def getDatasetWithID(self,datasetID):
        if datasetID in self.datasetMap:
            return self.datasetMap[datasetID]['datasetSpec']
        return None



    # return dataset with datasetName
    def getDatasetWithName(self,datasetName):
        for tmpDatasetID,tmpDatasetVal in iteritems(self.datasetMap):
            if tmpDatasetVal['datasetSpec'].datasetName == datasetName:
                return tmpDatasetVal['datasetSpec']
        return None



    # reset used counters
    def resetUsedCounters(self):
        for tmpKey,tmpVal in iteritems(self.datasetMap):
            tmpVal['used'] = 0

    # checkpoint file usage
    def checkpoint_file_usage(self):
        for tmpKey, tmpVal in iteritems(self.datasetMap):
            self.file_checkpoints[tmpKey] = tmpVal['used']

    # rollback file usage
    def rollback_file_usage(self):
        for tmpKey, tmpVal in iteritems(self.datasetMap):
            tmpVal['used'] = self.file_checkpoints[tmpKey]

    # add site candidates
    def addSiteCandidate(self,siteCandidateSpec):
        self.siteCandidates[siteCandidateSpec.siteName] = siteCandidateSpec
        return



    # add site candidates
    def addSiteCandidateForJumbo(self,siteCandidateSpec):
        self.siteCandidatesJumbo[siteCandidateSpec.siteName] = siteCandidateSpec
        return



    # has candidate for jumbo jobs
    def hasCandidatesForJumbo(self):
        return len(self.siteCandidatesJumbo) > 0



    # get one site candidate randomly
    def getOneSiteCandidate(self, nSubChunks=0, ngSites=None, get_msg=False):
        retSiteCandidate = None
        if ngSites is None:
            ngSites = []
        ngSites = copy.copy(ngSites)
        # skip sites for distributed datasets
        """
        for tmpDatasetSpec in self.getDatasets():
            if tmpDatasetSpec.isDistributed():
                datasetUsage = self.datasetMap[tmpDatasetSpec.datasetID]
                if len(tmpDatasetSpec.Files) > datasetUsage['used']:
                    tmpFileSpec = tmpDatasetSpec.Files[datasetUsage['used']]
                    for siteCandidate in self.siteCandidates.values():
                        # skip if the first file is unavailable at the site
                        if not siteCandidate.isAvailableFile(tmpFileSpec):
                            ngSites.append(siteCandidate.siteName)
        """
        # check if to be bootstrapped
        siteCandidateList = list(self.siteCandidates.values())
        newSiteCandidateList = []
        for siteCandidate in siteCandidateList:
            if siteCandidate.weight == 0 and siteCandidate.siteName not in self.bootstrapped:
                newSiteCandidateList.append(siteCandidate)
        if newSiteCandidateList:
            retSiteCandidate = random.choice(newSiteCandidateList)
            self.bootstrapped.add(retSiteCandidate.siteName)
            retMsg = 'toBootstrapped={}'.format(len(newSiteCandidateList))
        else:
            # get total weight
            totalWeight = 0
            nNG = 0
            nOK = 0
            nBoosted = 0
            nFull = 0
            fullStr = ''
            for siteCandidate in siteCandidateList:
                # remove NG sites
                if siteCandidate.siteName in ngSites:
                    nNG += 1
                    continue
                # already bootstrapped
                if siteCandidate.weight == 0 and siteCandidate.siteName in self.bootstrapped:
                    nBoosted += 1
                    continue
                # skip incapable
                if not siteCandidate.can_accept_jobs():
                    nFull += 1
                    fullStr += '{}:{}/{} '.format(siteCandidate.siteName,
                                                 siteCandidate.nQueuedJobs,
                                                 siteCandidate.nRunningJobsCap)
                    continue
                totalWeight += siteCandidate.weight
                newSiteCandidateList.append(siteCandidate)
                nOK += 1
            siteCandidateList = newSiteCandidateList
            if fullStr:
                fullStr = " (skipped {})".format(fullStr[:-1])
            retMsg = 'OK={} NG={} bootstrapped={} Full={}{}'.format(nOK, nNG, nBoosted, nFull,
                                                                    fullStr)
            # empty
            if not siteCandidateList:
                if get_msg:
                    return None, retMsg
                return None
            # get random number
            rNumber = random.random() * totalWeight
            for siteCandidate in siteCandidateList:
                rNumber -= siteCandidate.weight
                if rNumber <= 0:
                    retSiteCandidate = siteCandidate
                    break
            # return something as a protection against precision of float
            if retSiteCandidate is None:
                retSiteCandidate = random.choice(siteCandidateList)
            # modify weight
            try:
                if retSiteCandidate.nQueuedJobs is not None and retSiteCandidate.nAssignedJobs is not None:
                    oldNumQueued = retSiteCandidate.nQueuedJobs
                    retSiteCandidate.nQueuedJobs += nSubChunks
                    newNumQueued = retSiteCandidate.nQueuedJobs
                    retSiteCandidate.nAssignedJobs += nSubChunks
                    siteCandidate.weight = siteCandidate.weight * float(oldNumQueued+1) / float(newNumQueued+1)
            except Exception:
                pass
        if get_msg:
            return retSiteCandidate, retMsg
        return retSiteCandidate



    # get sites for parallel execution
    def getParallelSites(self,nSites,nSubChunks,usedSites):
        newSiteCandidate = self.getOneSiteCandidate(nSubChunks,usedSites)
        if newSiteCandidate is not None:
            usedSites.append(newSiteCandidate.siteName)
            if nSites > len(usedSites):
                return self.getParallelSites(nSites,nSubChunks,usedSites)
        return ','.join(usedSites)



    # get one site for jumbo jobs
    def getOneSiteCandidateForJumbo(self,ngSites):
        # get total weight
        totalWeight = 0
        weightList  = []
        siteCandidateList = list(self.siteCandidatesJumbo.values())
        newSiteCandidateList = []
        for siteCandidate in siteCandidateList:
            # remove NG sites
            if siteCandidate.siteName in ngSites:
                continue
            totalWeight += siteCandidate.weight
            newSiteCandidateList.append(siteCandidate)
        siteCandidateList = newSiteCandidateList
        # empty
        if siteCandidateList == []:
            return None
        # get random number
        rNumber = random.random() * totalWeight
        for siteCandidate in siteCandidateList:
            rNumber -= siteCandidate.weight
            if rNumber <= 0:
                retSiteCandidate = siteCandidate
                break
        # return something as a protection against precision of float
        if retSiteCandidate is None:
            retSiteCandidate = random.choice(siteCandidateList)
        return retSiteCandidate



    # check if unused files/events remain
    def checkUnused(self):
        # master is undefined
        if self.masterIndexName is None:
            return False
        indexVal = self.datasetMap[self.masterIndexName]
        return indexVal['used'] < len(indexVal['datasetSpec'].Files)



    # get master used index
    def getMasterUsedIndex(self):
        # master is undefined
        if self.masterIndexName is None:
            return 0
        indexVal = self.datasetMap[self.masterIndexName]
        return indexVal['used']



    # get num of files in master
    def getNumFilesInMaster(self):
        # master is undefined
        if self.masterIndexName is None:
            return 0
        indexVal = self.datasetMap[self.masterIndexName]
        return len(indexVal['datasetSpec'].Files)



    # check if secondary datasets use event ratios
    def useEventRatioForSec(self):
        for datasetSpec in self.secondaryDatasetList:
            if datasetSpec.getEventRatio() is not None:
                return True
        return False



    # get maximum size of atomic subchunk
    def getMaxAtomSize(self,effectiveSize=False,getNumEvents=False):
        # number of files per job if defined
        if not self.isMerging:
            nFilesPerJob = self.taskSpec.getNumFilesPerJob()
        else:
            nFilesPerJob = self.taskSpec.getNumFilesPerMergeJob()
        nEventsPerJob = None
        if nFilesPerJob is None:
            # number of events per job
            if not self.isMerging:
                nEventsPerJob = self.taskSpec.getNumEventsPerJob()
            else:
                nEventsPerJob = self.taskSpec.getNumEventsPerMergeJob()
            if nEventsPerJob is None:
                nFilesPerJob = 1
        # grouping with boundaryID
        useBoundary = self.taskSpec.useGroupWithBoundaryID()
        # LB
        respectLB = self.taskSpec.respectLumiblock()
        maxAtomSize = 0
        while True:
            if not self.isMerging:
                maxNumFiles = self.taskSpec.getMaxNumFilesPerJob()
            else:
                maxNumFiles = self.taskSpec.getMaxNumFilesPerMergeJob()
            # get one subchunk
            subChunk = self.getSubChunk(None,nFilesPerJob=nFilesPerJob,
                                        nEventsPerJob=nEventsPerJob,
                                        useBoundary=useBoundary,
                                        respectLB=respectLB,
                                        maxNumFiles=maxNumFiles)
            if subChunk is None:
                break
            # get size
            tmpAtomSize = 0
            for tmpDatasetSpec,tmpFileSpecList in subChunk:
                if (effectiveSize or getNumEvents) and not tmpDatasetSpec.isMaster():
                    continue
                for tmpFileSpec in tmpFileSpecList:
                    if effectiveSize:
                        tmpAtomSize += JediCoreUtils.getEffectiveFileSize(tmpFileSpec.fsize,tmpFileSpec.startEvent,
                                                                          tmpFileSpec.endEvent,tmpFileSpec.nEvents)
                    elif getNumEvents:
                        tmpAtomSize += tmpFileSpec.getEffectiveNumEvents()
                    else:
                        tmpAtomSize += tmpFileSpec.fsize
            if maxAtomSize < tmpAtomSize:
                maxAtomSize = tmpAtomSize
        # reset counters
        self.resetUsedCounters()
        # return
        return maxAtomSize



    # use scout
    def useScout(self):
        if self.masterDataset is not None and self.useScoutFlag is not None:
            return self.useScoutFlag
        if self.masterDataset is not None and \
                self.masterDataset.nFiles > self.masterDataset.nFilesToBeUsed:
            return True
        return False



    # set use scout
    def setUseScout(self,useScoutFlag):
        self.useScoutFlag = useScoutFlag



    # get preassigned site
    def getPreassignedSite(self):
        if self.masterDataset is not None:
            return self.masterDataset.site
        return None



    # get max output size
    def getOutSize(self,outSizeMap):
        values = list(outSizeMap.values())
        values.sort()
        try:
            return values[-1]
        except Exception:
            return 0

    # get value with unit
    def get_value_str(self, value):
        try:
            return '{}MB'.format(int(math.ceil(value/1024/1024)))
        except Exception:
            return None

    # get subchunk with a selection criteria
    def getSubChunk(self,siteName,maxNumFiles=None,maxSize=None,
                    sizeGradients=0,sizeIntercepts=0,
                    nFilesPerJob=None,multiplicand=1,
                    walltimeGradient=0,maxWalltime=0,
                    nEventsPerJob=None,useBoundary=None,
                    sizeGradientsPerInSize=None,
                    maxOutSize=None,
                    coreCount=1,
                    respectLB=False,
                    corePower=None,
                    dynNumEvents=False,
                    maxNumEventRanges=None,
                    multiplicity=None,
                    splitByFields=None,
                    tmpLog=None,
                    useDirectIO=False,
                    maxDiskSize=None,
                    enableLog=False):
        # check if there are unused files/events
        if not self.checkUnused():
            return None
        # protection against unreasonable values
        if nFilesPerJob == 0:
            nFilesPerJob = None
        if nEventsPerJob == 0:
            nEventsPerJob = None
        # set default max number of files
        if maxNumFiles is None:
            maxNumFiles = 200
        # set default max number of event ranges
        if maxNumEventRanges is None:
            maxNumEventRanges = 20
        # set default max size
        if maxSize is None and nFilesPerJob is None and nEventsPerJob is None:
            # 20 GB at most by default
            maxSize = 20 * 1024 * 1024 * 1024
        # set default output size
        minOutSize = self.defaultOutputSize
        # set default max number of events
        maxNumEvents = None
        # ignore negative walltime gradient
        if walltimeGradient is None or walltimeGradient < 0:
            walltimeGradient = 0
        # overwrite parameters when nFiles/EventsPerJob is used
        if nFilesPerJob is not None and not dynNumEvents:
            maxNumFiles  = nFilesPerJob
            if not respectLB:
                multiplicand = nFilesPerJob
        if nEventsPerJob is not None:
            maxNumEvents = nEventsPerJob
        # split with boundayID
        splitWithBoundaryID = False
        if useBoundary is not None:
            splitWithBoundaryID = True
            if useBoundary['inSplit'] == 2:
                # unset max values to split only with boundaryID
                maxNumFiles = None
                maxSize = None
                maxWalltime = 0
                maxNumEvents = None
                multiplicand = 1
        # get site when splitting per site
        if siteName is not None:
            siteCandidate = self.siteCandidates[siteName]
        # use event ratios
        useEventRatio = self.useEventRatioForSec()
        # start splitting
        inputNumFiles  = 0
        inputNumEvents = 0
        fileSize       = 0
        firstLoop      = True
        firstMaster    = True
        inputFileMap   = {}
        expWalltime    = 0
        nextStartEvent = None
        boundaryID     = None
        newBoundaryID  = False
        eventJump      = False
        nSecFilesMap   = {}
        nSecEventsMap  = {}
        numMaster      = 0
        masterSize     = 0
        outSizeMap     = {}
        lumiBlockNr    = None
        newLumiBlockNr = False
        siteAvailable  = True
        inputFileSet   = set()
        fieldStr       = None
        diskSize       = 0
        dumpStr = ''
        while (maxNumFiles is None or (not dynNumEvents and inputNumFiles <= maxNumFiles) or \
                   (dynNumEvents and len(inputFileSet) <= maxNumFiles and inputNumFiles <= maxNumEventRanges)) \
                and (maxSize is None or (maxSize is not None and fileSize <= maxSize)) \
                and (maxWalltime is None or maxWalltime <= 0 or expWalltime <= maxWalltime) \
                and (maxNumEvents is None or (maxNumEvents is not None and inputNumEvents <= maxNumEvents)) \
                and (maxOutSize is None or self.getOutSize(outSizeMap) <= maxOutSize) \
                and (maxDiskSize is None or diskSize <= maxDiskSize):
            # get one file (or one file group for MP) from master
            datasetUsage = self.datasetMap[self.masterDataset.datasetID]
            if self.masterDataset.datasetID not in outSizeMap:
                outSizeMap[self.masterDataset.datasetID] = 0
            boundaryIDs = set()
            primaryHasEvents = False
            for tmpFileSpec in self.masterDataset.Files[datasetUsage['used']:datasetUsage['used']+multiplicand]:
                # check start event to keep continuity
                if (maxNumEvents is not None or dynNumEvents) and tmpFileSpec.startEvent is not None:
                    if nextStartEvent is not None and nextStartEvent != tmpFileSpec.startEvent:
                        eventJump = True
                        break
                # check boundaryID
                if splitWithBoundaryID and boundaryID is not None and boundaryID != tmpFileSpec.boundaryID \
                        and useBoundary['inSplit'] != 3:
                    newBoundaryID = True
                    break
                # check LB
                if respectLB and lumiBlockNr is not None and lumiBlockNr != tmpFileSpec.lumiBlockNr:
                    newLumiBlockNr = True
                    break
                # check field
                if splitByFields is not None:
                    tmpFieldStr = tmpFileSpec.extractFieldsStr(splitByFields)
                    if fieldStr is None:
                        fieldStr = tmpFieldStr
                    elif tmpFieldStr != fieldStr:
                        newBoundaryID = True
                        break
                # check for distributed datasets
                #if self.masterDataset.isDistributed() and siteName is not None and \
                #        not siteCandidate.isAvailableFile(tmpFileSpec):
                #    siteAvailable = False
                #    break
                if self.masterDataset.datasetID not in inputFileMap:
                    inputFileMap[self.masterDataset.datasetID] = []
                inputFileMap[self.masterDataset.datasetID].append(tmpFileSpec)
                inputFileSet.add(tmpFileSpec.lfn)
                datasetUsage['used'] += 1
                numMaster += 1
                # get effective file size
                effectiveFsize = JediCoreUtils.getEffectiveFileSize(tmpFileSpec.fsize,tmpFileSpec.startEvent,
                                                                    tmpFileSpec.endEvent,tmpFileSpec.nEvents)
                # get num of events
                effectiveNumEvents = tmpFileSpec.getEffectiveNumEvents()
                # sum
                inputNumFiles += 1
                if self.taskSpec.outputScaleWithEvents():
                    tmpOutSize = long(sizeGradients * effectiveNumEvents)
                    fileSize += tmpOutSize
                    diskSize += tmpOutSize
                    if not dynNumEvents or tmpFileSpec.lfn not in inputFileSet:
                        fileSize += long(tmpFileSpec.fsize)
                        masterSize += long(tmpFileSpec.fsize)
                        if not useDirectIO:
                            diskSize += long(tmpFileSpec.fsize)
                    outSizeMap[self.masterDataset.datasetID] += long(sizeGradients * effectiveNumEvents)
                else:
                    tmpOutSize = long(sizeGradients * effectiveFsize)
                    fileSize += tmpOutSize
                    diskSize += tmpOutSize
                    if not dynNumEvents or tmpFileSpec.lfn not in inputFileSet:
                        fileSize += long(tmpFileSpec.fsize)
                        masterSize += long(tmpFileSpec.fsize)
                        if not useDirectIO:
                            diskSize += long(tmpFileSpec.fsize)
                    outSizeMap[self.masterDataset.datasetID] += long(sizeGradients * effectiveFsize)
                if sizeGradientsPerInSize is not None:
                    tmpOutSize = long(effectiveFsize * sizeGradientsPerInSize)
                    fileSize += tmpOutSize
                    diskSize += tmpOutSize
                    outSizeMap[self.masterDataset.datasetID] += long(effectiveFsize * sizeGradientsPerInSize)
                # sum offset only for the first master
                if firstMaster:
                    fileSize += sizeIntercepts
                # walltime
                if self.taskSpec.useHS06():
                    if firstMaster:
                        expWalltime += self.taskSpec.baseWalltime
                    tmpExpWalltime = walltimeGradient * effectiveNumEvents / float(coreCount)
                    if corePower not in [None,0]:
                        tmpExpWalltime /= corePower
                    if self.taskSpec.cpuEfficiency == 0:
                        tmpExpWalltime = 0
                    else:
                        tmpExpWalltime /= float(self.taskSpec.cpuEfficiency)/100.0
                    if multiplicity is not None:
                        tmpExpWalltime /= float(multiplicity)
                    expWalltime += long(tmpExpWalltime)
                else:
                    tmpExpWalltime = walltimeGradient * effectiveFsize / float(coreCount)
                    if multiplicity is not None:
                        tmpExpWalltime /= float(multiplicity)
                    expWalltime += long(tmpExpWalltime)
                # the number of events
                if (maxNumEvents is not None or useEventRatio) and tmpFileSpec.startEvent is not None and tmpFileSpec.endEvent is not None:
                    primaryHasEvents = True
                    inputNumEvents += (tmpFileSpec.endEvent - tmpFileSpec.startEvent + 1)
                    # set next start event
                    nextStartEvent = tmpFileSpec.endEvent + 1
                    if nextStartEvent == tmpFileSpec.nEvents:
                        nextStartEvent = 0
                # boundaryID
                if splitWithBoundaryID:
                    boundaryID = tmpFileSpec.boundaryID
                    if boundaryID not in boundaryIDs:
                        boundaryIDs.add(boundaryID)
                # LB
                if respectLB:
                    lumiBlockNr = tmpFileSpec.lumiBlockNr
                firstMaster = False
            # get files from secondaries
            firstSecondary = True
            for datasetSpec in self.secondaryDatasetList:
                if datasetSpec.datasetID not in outSizeMap:
                    outSizeMap[datasetSpec.datasetID] = 0
                if datasetSpec.isNoSplit():
                    # every job uses dataset without splitting
                    if firstLoop:
                        datasetUsage = self.datasetMap[datasetSpec.datasetID]
                        for tmpFileSpec in datasetSpec.Files:
                            if datasetSpec.datasetID not in inputFileMap:
                                inputFileMap[datasetSpec.datasetID] = []
                            inputFileMap[datasetSpec.datasetID].append(tmpFileSpec)
                            # sum
                            fileSize += tmpFileSpec.fsize
                            if not useDirectIO:
                                diskSize += tmpFileSpec.fsize
                            if sizeGradientsPerInSize is not None:
                                tmpOutSize = (tmpFileSpec.fsize * sizeGradientsPerInSize)
                                fileSize += tmpOutSize
                                diskSize += tmpOutSize
                                outSizeMap[datasetSpec.datasetID] += (tmpFileSpec.fsize * sizeGradientsPerInSize)
                            datasetUsage['used'] += 1
                else:
                    if datasetSpec.datasetID not in nSecFilesMap:
                        nSecFilesMap[datasetSpec.datasetID] = 0
                    # get number of files to be used for the secondary
                    nSecondary = datasetSpec.getNumFilesPerJob()
                    if nSecondary is not None and firstLoop is False:
                        # read files only in the first bunch when number of files per job is specified
                        continue
                    if nSecondary is None:
                        nSecondary = datasetSpec.getNumMultByRatio(numMaster) - nSecFilesMap[datasetSpec.datasetID]
                        if (datasetSpec.getEventRatio() is not None and inputNumEvents > 0) or (splitWithBoundaryID and useBoundary['inSplit'] != 3):
                            # set large number to get all associated secondary files
                            nSecondary = 10000
                    datasetUsage = self.datasetMap[datasetSpec.datasetID]
                    # reset nUsed
                    if datasetSpec.isReusable() and datasetUsage['used']+nSecondary > len(datasetSpec.Files):
                        datasetUsage['used'] = 0
                    for tmpFileSpec in datasetSpec.Files[datasetUsage['used']:datasetUsage['used']+nSecondary]:
                        # check boundaryID
                        if (splitWithBoundaryID or (useBoundary is not None and useBoundary['inSplit'] == 3 and datasetSpec.getRatioToMaster() > 1)) \
                                and boundaryID is not None and \
                                not (boundaryID == tmpFileSpec.boundaryID or tmpFileSpec.boundaryID in boundaryIDs):
                            break
                        # check for distributed datasets
                        #if datasetSpec.isDistributed() and siteName is not None and \
                        #        not siteCandidate.isAvailableFile(tmpFileSpec):
                        #    break
                        # check ratio
                        if datasetSpec.datasetID not in nSecEventsMap:
                            nSecEventsMap[datasetSpec.datasetID] = 0
                        if datasetSpec.getEventRatio() is not None and inputNumEvents > 0:
                            if float(nSecEventsMap[datasetSpec.datasetID]) / float(inputNumEvents) >= datasetSpec.getEventRatio():
                                break
                        if datasetSpec.datasetID not in inputFileMap:
                            inputFileMap[datasetSpec.datasetID] = []
                        inputFileMap[datasetSpec.datasetID].append(tmpFileSpec)
                        # sum
                        fileSize += tmpFileSpec.fsize
                        if not useDirectIO:
                            diskSize += tmpFileSpec.fsize
                        if sizeGradientsPerInSize is not None:
                            tmpOutSize = (tmpFileSpec.fsize * sizeGradientsPerInSize)
                            fileSize += tmpOutSize
                            diskSize += tmpOutSize
                            outSizeMap[datasetSpec.datasetID] += (tmpFileSpec.fsize * sizeGradientsPerInSize)
                        datasetUsage['used'] += 1
                        nSecFilesMap[datasetSpec.datasetID] += 1
                        # the number of events
                        if firstSecondary and maxNumEvents is not None and not primaryHasEvents:
                            if tmpFileSpec.startEvent is not None and tmpFileSpec.endEvent is not None:
                                inputNumEvents += (tmpFileSpec.endEvent - tmpFileSpec.startEvent + 1)
                            elif tmpFileSpec.nEvents is not None:
                                inputNumEvents += tmpFileSpec.nEvents
                        if tmpFileSpec.nEvents is not None:
                            nSecEventsMap[datasetSpec.datasetID] += tmpFileSpec.nEvents
                    # use only the first secondary
                    firstSecondary = False
            # unset first loop flag
            firstLoop = False
            # check if there are unused files/evets
            if not self.checkUnused():
                dumpStr = 'no more files'
                break
            # break if nFilesPerJob is used as multiplicand
            if nFilesPerJob is not None and not respectLB:
                dumpStr = 'nFilesPerJob specified'
                break
            # boundaryID is changed
            if newBoundaryID:
                dumpStr = 'new BoundaryID'
                break
            # LB is changed
            if newLumiBlockNr:
                dumpStr = 'new LB'
                break
            # event jump
            if eventJump:
                dumpStr = 'event jump'
                break
            # distributed files are unavailable
            if not siteAvailable:
                dumpStr = 'distributed files are unavailable'
                break
            primaryHasEvents = False
            # check master in the next loop
            datasetUsage = self.datasetMap[self.masterDataset.datasetID]
            newInputNumFiles  = inputNumFiles
            newInputNumEvents = inputNumEvents
            newFileSize       = fileSize
            newExpWalltime    = expWalltime
            newNextStartEvent = nextStartEvent
            newNumMaster      = numMaster
            terminateFlag     = False
            newOutSizeMap     = copy.copy(outSizeMap)
            newBoundaryIDs    = set()
            newInputFileSet   = copy.copy(inputFileSet)
            newDiskSize       = diskSize
            new_nSecEventsMap = copy.copy(nSecEventsMap)
            if self.masterDataset.datasetID not in newOutSizeMap:
                newOutSizeMap[self.masterDataset.datasetID] = 0
            for tmpFileSpec in self.masterDataset.Files[datasetUsage['used']:datasetUsage['used']+multiplicand]:
                # check continuity of event
                if maxNumEvents is not None and tmpFileSpec.startEvent is not None and tmpFileSpec.endEvent is not None:
                    primaryHasEvents = True
                    newInputNumEvents += (tmpFileSpec.endEvent - tmpFileSpec.startEvent + 1)
                    # continuity of event is broken
                    if newNextStartEvent is not None and newNextStartEvent != tmpFileSpec.startEvent:
                        # no files in the next loop
                        if newInputNumFiles == 0:
                            dumpStr = 'no files with continuous events in the next loop'
                            terminateFlag = True
                        break
                    newNextStartEvent = tmpFileSpec.endEvent + 1
                # check boundary
                if splitWithBoundaryID and boundaryID is not None and boundaryID != tmpFileSpec.boundaryID \
                        and useBoundary['inSplit'] != 3:
                    # no files in the next loop
                    if newInputNumFiles == 0:
                        dumpStr = 'no files with the same BoundaryID in the next loop'
                        terminateFlag = True
                    break
                # check LB
                if respectLB and lumiBlockNr is not None and lumiBlockNr != tmpFileSpec.lumiBlockNr:
                    # no files in the next loop
                    if newInputNumFiles == 0:
                        dumpStr = 'no files with the same LB in the next loop'
                        terminateFlag = True
                    break
                # check field
                if splitByFields is not None:
                    tmpFieldStr = tmpFileSpec.extractFieldsStr(splitByFields)
                    if tmpFieldStr != fieldStr:
                        # no files in the next loop
                        if newInputNumFiles == 0:
                            dumpStr = 'no files with the same LFN field in the next loop'
                            terminateFlag = True
                        break
                # check for distributed datasets
                #if self.masterDataset.isDistributed() and siteName is not None and \
                #        not siteCandidate.isAvailableFile(tmpFileSpec):
                #    # no files in the next loop
                #    if newInputNumFiles == 0:
                #        terminateFlag = True
                #    break
                # get effective file size
                effectiveFsize = JediCoreUtils.getEffectiveFileSize(tmpFileSpec.fsize,tmpFileSpec.startEvent,
                                                                    tmpFileSpec.endEvent,tmpFileSpec.nEvents)
                # get num of events
                effectiveNumEvents = tmpFileSpec.getEffectiveNumEvents()
                newInputNumFiles += 1
                newNumMaster += 1
                newInputFileSet.add(tmpFileSpec.lfn)
                if self.taskSpec.outputScaleWithEvents():
                    tmpOutSize = long(sizeGradients * effectiveNumEvents)
                    newFileSize += tmpOutSize
                    newDiskSize += tmpOutSize
                    if not dynNumEvents or tmpFileSpec.lfn not in inputFileSet:
                        newFileSize += long(tmpFileSpec.fsize)
                        if not useDirectIO:
                            newDiskSize += long(tmpFileSpec.fsize)
                    newOutSizeMap[self.masterDataset.datasetID] += long(sizeGradients * effectiveNumEvents)
                else:
                    tmpOutSize = long(sizeGradients * effectiveFsize)
                    newFileSize += tmpOutSize
                    newDiskSize += tmpOutSize
                    if not dynNumEvents or tmpFileSpec.lfn not in inputFileSet:
                        newFileSize += long(tmpFileSpec.fsize)
                        if not useDirectIO:
                            newDiskSize += long(tmpFileSpec.fsize)
                    newOutSizeMap[self.masterDataset.datasetID] += long(sizeGradients * effectiveFsize)
                if sizeGradientsPerInSize is not None:
                    tmpOutSize = long(effectiveFsize * sizeGradientsPerInSize)
                    newFileSize += tmpOutSize
                    newDiskSize += tmpOutSize
                    newOutSizeMap[self.masterDataset.datasetID] += long(effectiveFsize * sizeGradientsPerInSize)
                if self.taskSpec.useHS06():
                    tmpExpWalltime = walltimeGradient * effectiveNumEvents / float(coreCount)
                    if corePower not in [None,0]:
                        tmpExpWalltime /= corePower
                    if self.taskSpec.cpuEfficiency == 0:
                        tmpExpWalltime = 0
                    else:
                        tmpExpWalltime /= float(self.taskSpec.cpuEfficiency)/100.0
                    if multiplicity is not None:
                        tmpExpWalltime /= float(multiplicity)
                    newExpWalltime += long(tmpExpWalltime)
                else:
                    tmpExpWalltime = walltimeGradient * effectiveFsize / float(coreCount)
                    if multiplicity is not None:
                        tmpExpWalltime /= float(multiplicity)
                    newExpWalltime += long(tmpExpWalltime)
                # boundaryID
                if splitWithBoundaryID:
                    newBoundaryIDs.add(tmpFileSpec.boundaryID)
            # check secondaries
            firstSecondary = True
            newSecMap = {}
            for datasetSpec in self.secondaryDatasetList:
                if datasetSpec.datasetID not in newOutSizeMap:
                    newOutSizeMap[datasetSpec.datasetID] = 0
                if not datasetSpec.isNoSplit() and datasetSpec.getNumFilesPerJob() is None:
                    # check boundaryID
                    if splitWithBoundaryID and boundaryID is not None and boundaryID != tmpFileSpec.boundaryID \
                            and useBoundary['inSplit'] != 3:
                        break
                    newSecMap.setdefault(datasetSpec.datasetID, {'in_size': 0, 'out_size': 0})
                    newNumSecondary = datasetSpec.getNumMultByRatio(newNumMaster) - nSecFilesMap[datasetSpec.datasetID]
                    if newNumSecondary < 0:
                        newNumSecondary = 0
                    newSecMap[datasetSpec.datasetID]['nSec'] = newNumSecondary
                    newSecMap[datasetSpec.datasetID]['nSecReal'] = 0
                    datasetUsage = self.datasetMap[datasetSpec.datasetID]
                    for tmpFileSpec in datasetSpec.Files[datasetUsage['used']:datasetUsage['used']+newNumSecondary]:
                        # check boundaryID
                        if splitWithBoundaryID and boundaryID is not None and boundaryID != tmpFileSpec.boundaryID \
                                and tmpFileSpec.boundaryID not in boundaryIDs and tmpFileSpec.boundaryID not in newBoundaryIDs:
                            break
                        # check ratio
                        if datasetSpec.getEventRatio() is not None and newInputNumEvents > 0:
                            if float(new_nSecEventsMap[datasetSpec.datasetID]) / float(newInputNumEvents) \
                                    >= datasetSpec.getEventRatio():
                                break
                        newFileSize += tmpFileSpec.fsize
                        newSecMap[datasetSpec.datasetID]['in_size'] += tmpFileSpec.fsize
                        newSecMap[datasetSpec.datasetID]['nSecReal'] += 1
                        if not useDirectIO:
                            newDiskSize += tmpFileSpec.fsize
                        if sizeGradientsPerInSize is not None:
                            tmpOutSize = (tmpFileSpec.fsize * sizeGradientsPerInSize)
                            newFileSize += tmpOutSize
                            newDiskSize += tmpOutSize
                            newOutSizeMap[datasetSpec.datasetID] += tmpOutSize
                            newSecMap[datasetSpec.datasetID]['out_size'] += tmpOutSize
                        # the number of events
                        if firstSecondary and maxNumEvents is not None and not primaryHasEvents:
                            if tmpFileSpec.startEvent is not None and tmpFileSpec.endEvent is not None:
                                newInputNumEvents += (tmpFileSpec.endEvent - tmpFileSpec.startEvent + 1)
                            elif tmpFileSpec.nEvents is not None:
                                newInputNumEvents += tmpFileSpec.nEvents
                        if tmpFileSpec.nEvents is not None:
                            new_nSecEventsMap[datasetSpec.datasetID] += tmpFileSpec.nEvents
                    firstSecondary = False
            # termination
            if terminateFlag:
                break
            # check
            newOutSize = self.getOutSize(newOutSizeMap)
            if (maxNumFiles is not None and ((not dynNumEvents and newInputNumFiles > maxNumFiles) \
                                             or (dynNumEvents and (len(newInputFileSet) > maxNumFiles or newInputNumFiles > maxNumEventRanges)))) \
                    or (maxSize is not None and newFileSize > maxSize) \
                    or (maxSize is not None and newOutSize < minOutSize and maxSize-minOutSize < newFileSize-newOutSize) \
                    or (maxWalltime is not None and 0 < maxWalltime < newExpWalltime) \
                    or (maxNumEvents is not None and newInputNumEvents > maxNumEvents) \
                    or (maxOutSize is not None and self.getOutSize(newOutSizeMap) > maxOutSize) \
                    or (maxDiskSize is not None and newDiskSize > maxDiskSize):
                dumpStr = 'check for the next loop. '
                if maxNumFiles is not None and (not dynNumEvents and newInputNumFiles > maxNumFiles):
                    dumpStr += 'maxNumFiles exceeds maxNumFiles={} inputNumFiles={} newInputNumFiles={}. '.format(
                        maxNumFiles, inputNumFiles, newInputNumFiles)
                if maxSize is not None and newFileSize > maxSize:
                    dumpStr += 'maxSize exceeds maxSize={} fileSize={} masterSize={} newFileSize={} newSecMap={}. '.format(
                        self.get_value_str(maxSize), self.get_value_str(fileSize), self.get_value_str(masterSize),
                        self.get_value_str(newFileSize), str(newSecMap))
                if maxSize is not None and newOutSize < minOutSize and maxSize-minOutSize < newFileSize-newOutSize:
                    dumpStr += 'maxSize exceeds with outSize maxSize={} minOutSize={} fileSize={} newFileSize={} newOutSize={}. '.\
                        format(self.get_value_str(maxSize), self.get_value_str(minOutSize), self.get_value_str(fileSize),
                               self.get_value_str(newFileSize), self.get_value_str(newOutSize))
                if maxWalltime is not None and 0 < maxWalltime < newExpWalltime:
                    dumpStr += 'maxWalltime exceeds maxWalltime={} expWalltime={} newExpWalltime={}. '.format(
                        maxWalltime, expWalltime, newExpWalltime)
                if maxNumEvents is not None and newInputNumEvents > maxNumEvents:
                    dumpStr += 'maxNumEvents exceeds maxNumEvents={} inputNumEvents={} newInputNumEvents={}. '.format(
                        maxNumEvents, inputNumEvents, newInputNumEvents)
                if maxOutSize is not None and self.getOutSize(newOutSizeMap) > maxOutSize:
                    dumpStr += 'maxOutSize exceeds maxOutSize={} getOutSize(newOutSizeMap)={}. '.format(
                        self.get_value_str(maxOutSize), self.get_value_str(self.getOutSize(newOutSizeMap)))
                if maxDiskSize is not None and newDiskSize > maxDiskSize:
                    dumpStr += 'maxDiskSize exceeds maxDiskSize={} diskSize={} newDiskSize={}. '.format(
                        self.get_value_str(maxDiskSize), self.get_value_str(diskSize), self.get_value_str(newDiskSize))
                break
        # reset nUsed for repeated datasets
        for tmpDatasetID,datasetUsage in iteritems(self.datasetMap):
            tmpDatasetSpec = datasetUsage['datasetSpec']
            if tmpDatasetSpec.isRepeated():
                if len(tmpDatasetSpec.Files) > 0:
                    datasetUsage['used'] %= len(tmpDatasetSpec.Files)
        # make copy to return
        returnList = []
        for tmpDatasetID,inputFileList in iteritems(inputFileMap):
            tmpRetList = []
            for tmpFileSpec in inputFileList:
                # split par site or get atomic subchunk
                if siteName is not None:
                    # make copy to individually set locality
                    newFileSpec = copy.copy(tmpFileSpec)
                    # set locality
                    newFileSpec.locality = siteCandidate.getFileLocality(tmpFileSpec)
                    if newFileSpec.locality == 'remote':
                        newFileSpec.sourceName = siteCandidate.remoteSource
                    # append
                    tmpRetList.append(newFileSpec)
                else:
                    # getting atomic subchunk
                    tmpRetList.append(tmpFileSpec)
            # add to return map
            tmpDatasetSpec = self.getDatasetWithID(tmpDatasetID)
            returnList.append((tmpDatasetSpec,tmpRetList))
        # dump only problematic splitting
        if enableLog and tmpLog and returnList:
            if maxNumEvents is not None and inputNumEvents < maxNumEvents:
                tmpLog.debug('not enough events {}>{} at {}: numMaster={}. {}'.format(maxNumEvents, inputNumEvents,
                                                                                      siteName, numMaster, dumpStr))
        # return
        return returnList



    # check if master is mutable
    def isMutableMaster(self):
        if self.masterDataset is not None and self.masterDataset.state == 'mutable':
            return True
        return False



    # figure out if output will go through express stream
    def isExpress(self):
        if self.taskSpec.processingType == 'urgent' or self.taskSpec.currentPriority > 1000:
            return True

        return False




    # skip unavailable files in distributed datasets
    def skipUnavailableFiles(self):
        # skip files if no candiate have them
        nSkip = 0
        for tmpDatasetSpec in self.getDatasets():
            if tmpDatasetSpec.isDistributed():
                datasetUsage = self.datasetMap[tmpDatasetSpec.datasetID]
                while len(tmpDatasetSpec.Files) > datasetUsage['used']:
                    tmpFileSpec = tmpDatasetSpec.Files[datasetUsage['used']]
                    isOK = False
                    for siteCandidate in self.siteCandidates.values():
                        if siteCandidate.isAvailableFile(tmpFileSpec):
                            isOK = True
                            break
                    if isOK:
                        break
                    # skip and check the next
                    datasetUsage['used'] += 1
                    nSkip += 1
        return nSkip



    # get max ramCount
    def getMaxRamCount(self):
        if self.isMerging:
            return max(self.taskSpec.mergeRamCount, self.ramCount) if self.taskSpec.mergeRamCount else self.ramCount
        else:
            return max(self.taskSpec.ramCount, self.ramCount) if self.taskSpec.ramCount else self.ramCount


    # get site candidate
    def getSiteCandidate(self, name):
        if name in self.siteCandidates:
            return self.siteCandidates[name]
        return None

    # get list of candidate names
    def get_candidate_names(self):
        return list(self.siteCandidates.keys())

    # update number of queued jobs
    def update_n_queue(self, live_counter):
        sites = []
        for siteCandidate in self.siteCandidates.values():
            if live_counter is not None:
                n = live_counter.get(siteCandidate.siteName)
                if n > 0:
                    siteCandidate.nQueuedJobs += n
                    sites.append(siteCandidate.siteName)
        return ','.join(sites)



import os
import sys
import time
import socket
import datetime

from pandajedi.jedicore.ThreadUtils import ListWithLock,ThreadPool,WorkerThread
from pandajedi.jedicore import Interaction
from pandajedi.jedicore.MsgWrapper import MsgWrapper
from pandajedi.jedicore.FactoryBase import FactoryBase
from .JediKnight import JediKnight
from pandajedi.jediconfig import jedi_config


# logger
from pandacommon.pandalogger.PandaLogger import PandaLogger
logger = PandaLogger().getLogger(__name__.split('.')[-1])


# worker class to do post-processing
class PostProcessor (JediKnight,FactoryBase):

    # constructor
    def __init__(self,commuChannel,taskBufferIF,ddmIF,vos,prodSourceLabels):
        self.vos = self.parseInit(vos)
        self.prodSourceLabels = self.parseInit(prodSourceLabels)
        self.pid = '{0}-{1}-post'.format(socket.getfqdn().split('.')[0],os.getpid())
        JediKnight.__init__(self,commuChannel,taskBufferIF,ddmIF,logger)
        FactoryBase.__init__(self,self.vos,self.prodSourceLabels,logger,
                             jedi_config.postprocessor.modConfig)



    # main
    def start(self):
        # start base classes
        JediKnight.start(self)
        FactoryBase.initializeMods(self,self.taskBufferIF,self.ddmIF)
        # go into main loop
        while True:
            startTime = datetime.datetime.utcnow()
            try:
                # get logger
                tmpLog = MsgWrapper(logger)
                tmpLog.info('start')
                # loop over all vos
                for vo in self.vos:
                    # loop over all sourceLabels
                    for prodSourceLabel in self.prodSourceLabels:
                        # prepare tasks to be finished
                        tmpLog.info('preparing tasks to be finished for vo={0} label={1}'.format(vo,prodSourceLabel))
                        tmpRet = self.taskBufferIF.prepareTasksToBeFinished_JEDI(vo,prodSourceLabel,
                                                                                 jedi_config.postprocessor.nTasks,
                                                                                 pid=self.pid)
                        if tmpRet is None:
                            # failed
                            tmpLog.error('failed to prepare tasks')
                        # get tasks to be finished
                        tmpLog.info('getting tasks to be finished') 
                        tmpList = self.taskBufferIF.getTasksToBeFinished_JEDI(vo,prodSourceLabel,self.pid,
                                                                              jedi_config.postprocessor.nTasks)
                        if tmpList is None: 
                            # failed
                            tmpLog.error('failed to get tasks to be finished')
                        else:
                            tmpLog.info('got {0} tasks'.format(len(tmpList)))
                            # put to a locked list
                            taskList = ListWithLock(tmpList)
                            # make thread pool
                            threadPool = ThreadPool()
                            # make workers
                            nWorker = jedi_config.postprocessor.nWorkers
                            for iWorker in range(nWorker):
                                thr = PostProcessorThread(taskList,threadPool,
                                                          self.taskBufferIF,
                                                          self.ddmIF,
                                                          self)
                                thr.start()
                            # join
                            threadPool.join()
                tmpLog.info('done')
            except Exception:
                errtype,errvalue = sys.exc_info()[:2]
                tmpLog.error('failed in {0}.start() with {1} {2}'.format(self.__class__.__name__,errtype.__name__,errvalue))
            # sleep if needed
            loopCycle = 60
            timeDelta = datetime.datetime.utcnow() - startTime
            sleepPeriod = loopCycle - timeDelta.seconds
            if sleepPeriod > 0:
                time.sleep(sleepPeriod)



# thread for real worker
class PostProcessorThread (WorkerThread):

    # constructor
    def __init__(self,taskList,threadPool,taskbufferIF,ddmIF,implFactory):
        # initialize woker with no semaphore
        WorkerThread.__init__(self,None,threadPool,logger)
        # attributres
        self.taskList = taskList
        self.taskBufferIF = taskbufferIF
        self.ddmIF = ddmIF
        self.implFactory = implFactory


    # main
    def runImpl(self):
        while True:
            try:
                # get a part of list
                nTasks = 10
                taskList = self.taskList.get(nTasks)
                # no more datasets
                if len(taskList) == 0:
                    self.logger.debug('{0} terminating since no more items'.format(self.__class__.__name__))
                    return
                # loop over all tasks
                for taskSpec in taskList:
                    # make logger
                    tmpLog = MsgWrapper(self.logger,'<jediTaskID={0}>'.format(taskSpec.jediTaskID))
                    tmpLog.info('start')
                    tmpStat = Interaction.SC_SUCCEEDED
                    # get impl
                    impl = self.implFactory.instantiateImpl(taskSpec.vo,taskSpec.prodSourceLabel,None,
                                                            self.taskBufferIF,self.ddmIF)
                    if impl is None:
                        # post processor is undefined
                        tmpLog.error('post-processor is undefined for vo={0} sourceLabel={1}'.format(taskSpec.vo,taskSpec.prodSourceLabel))
                        tmpStat = Interaction.SC_FATAL
                    # execute    
                    if tmpStat == Interaction.SC_SUCCEEDED:
                        tmpLog.info('post-process with {0}'.format(impl.__class__.__name__))
                        try:
                            impl.doPostProcess(taskSpec,tmpLog)
                        except Exception:
                            errtype,errvalue = sys.exc_info()[:2]
                            tmpLog.error('doPostProcess failed with {0}:{1}'.format(errtype.__name__,errvalue))
                            tmpStat = Interaction.SC_FATAL
                    # done
                    if tmpStat == Interaction.SC_FATAL:
                        # task is broken
                        tmpErrStr = 'post-process failed'
                        tmpLog.error(tmpErrStr)
                        taskSpec.status = 'broken'
                        taskSpec.setErrDiag(tmpErrStr)
                        taskSpec.lockedBy = None
                        self.taskBufferIF.updateTask_JEDI(taskSpec,{'jediTaskID':taskSpec.jediTaskID})    
                    elif tmpStat == Interaction.SC_FAILED:
                        tmpErrStr = 'post processing failed'
                        taskSpec.setOnHold()
                        taskSpec.setErrDiag(tmpErrStr,True)
                        taskSpec.lockedBy = None
                        self.taskBufferIF.updateTask_JEDI(taskSpec,{'jediTaskID':taskSpec.jediTaskID})
                        tmpLog.info('set task_status={0} since {1}'.format(taskSpec.status,taskSpec.errorDialog))
                        continue
                    # final procedure
                    try:
                        impl.doFinalProcedure(taskSpec,tmpLog)
                    except Exception:
                        errtype,errvalue = sys.exc_info()[:2]
                        tmpLog.error('doFinalProcedure failed with {0}:{1}'.format(errtype.__name__,errvalue))
                    # done
                    tmpLog.info('done')
            except Exception:
                errtype,errvalue = sys.exc_info()[:2]
                logger.error('{0} failed in runImpl() with {1}:{2}'.format(self.__class__.__name__,errtype.__name__,errvalue))
        


########## launch 
                
def launcher(commuChannel,taskBufferIF,ddmIF,vos=None,prodSourceLabels=None):
    p = PostProcessor(commuChannel,taskBufferIF,ddmIF,vos,prodSourceLabels)
    p.start()
